#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "endian.h"

#ifdef IS_LITTLE_ENDIAN

void swab2(char *ptr, size_t nmemb)
{
    size_t offset;
    for (offset = 0; offset < nmemb * 2; offset += 2) {
	char a, b;
	a = ptr[offset];
	b = ptr[offset + 1];
	ptr[offset] = b;
	ptr[offset + 1] = a;
    }
}

void swab4(char *ptr, size_t nmemb)
{
    size_t offset;
    for (offset = 0; offset < nmemb * 4; offset += 4) {
	char a, b, c, d;
	a = ptr[offset];
	b = ptr[offset + 1];
	c = ptr[offset + 2];
	d = ptr[offset + 3];
	ptr[offset] = d;
	ptr[offset + 1] = c;
	ptr[offset + 2] = b;
	ptr[offset + 3] = a;
    }
}

void swab8(char *ptr, size_t nmemb)
{
    size_t offset;
    char buf[8];
    int idx;
    for (offset = 0; offset < nmemb * 8; offset += 8) {
	for (idx = 0; idx < 8; idx++)
	    buf[idx] = ptr[offset + idx];
	for (idx = 0; idx < 8; idx++)
	    ptr[offset + idx] = buf[8 - 1 - idx];
    }
}

void swab16(char *ptr, size_t nmemb)
{
    size_t offset;
    char buf[16];
    int idx;
    for (offset = 0; offset < nmemb * 16; offset += 16) {
	for (idx = 0; idx < 16; idx++)
	    buf[idx] = ptr[offset + idx];
	for (idx = 0; idx < 16; idx++)
	    ptr[offset + idx] = buf[16 - 1 - idx];
    }
}

# define SWAB2(ptr, nmemb) swab2(ptr, nmemb)
# define SWAB4(ptr, nmemb) swab4(ptr, nmemb)
# define SWAB8(ptr, nmemb) swab8(ptr, nmemb)
# define SWAB16(ptr, nmemb) swab16(ptr, nmemb)

#else				/* IS_LITTLE_ENDIAN */

# define SWAB2(ptr, nmemb) 1
# define SWAB4(ptr, nmemb) 1
# define SWAB8(ptr, nmemb) 1
# define SWAB16(ptr, nmemb) 1

#endif				/* IS_LITTLE_ENDIAN */

    size_t
fread_native(void *ptr, size_t size, size_t nmemb, FILE * stream)
{
    size_t r;
    r = fread(ptr, size, nmemb, stream);
#ifdef IS_LITTLE_ENDIAN
    switch (size) {
    case 1:
	;
	break;
    case 2:
	SWAB2(ptr, nmemb);
	break;
    case 4:
	SWAB4(ptr, nmemb);
	break;
    case 8:
	SWAB8(ptr, nmemb);
	break;
    case 16:
	SWAB16(ptr, nmemb);
	break;
    default:
	fprintf(stderr,  "fread_native(%p, invalid:%lu, %lu, %p)\n",
                ptr, (unsigned long)size, (unsigned long)nmemb, stream);
	return -1;
    }
#endif
    return r;
}

size_t fread_fmt(void *ptr, char *fmt, FILE * stream)
{
    char *p, *f, *f2;
    int conv;
    unsigned long count;
    size_t sum;

    sum = 0;
    p = (char*)ptr;
    for (f = fmt; *f; f++) {
	count = strtoul(f, &f2, 0);
	if (f2 == f) {
	    count = 1;
	    conv = *f2;
	} else if (f2 == NULL) {
	    conv = 'b';
	    /* this causes loop-break after f++ */
	    f = "*";
	} else {
	    conv = *f2;
	    f = f2;
	}
        if(conv == 'q' || conv == 'Q'){
            sum += 16 * count;
        } else if (conv == 'd' || conv == 'D' || conv == 'l'
                 || conv == 'L'){
            sum += 8 * count;
	} else if (conv == 'f' || conv == 'F' || conv == 'i'
		   || conv == 'I') {
            sum += 4 * count;
	} else if (conv == 'h' || conv == 'H') {
            sum +=  2 * count;
	} else {
            sum += 1 * count;
        }            
    }

    if(fread(p, sum, 1, stream) != 1){
        return 0;
    }
#ifndef IS_LITTLE_ENDIAN
    p += sum;
#else
    for (f = fmt; *f; f++) {
        count = strtoul(f, &f2, 0);
        if (f2 == f) {
            count = 1;
            conv = *f2;
        } else if (f2 == NULL) {
            conv = 'b';
            /* this causes loop-break after f++ */
            f = "*";
        } else {
            conv = *f2;
            f = f2;
        }
        
        if (conv == 'q' || conv == 'Q') {
            endian_swab16(p, count);
            p += count * 16;
        } else if (conv == 'd' || conv == 'D' || conv == 'l'
                   || conv == 'L') {
            endian_swab8(p, count);
            p += count * 8;
        } else if (conv == 'f' || conv == 'F' || conv == 'i'
                   || conv == 'I') {
            endian_swab4(p, count);
            p += count * 4;
        } else if (conv == 'h' || conv == 'H') {
            endian_swab2(p, count);
            p += count * 2;
        } else {
            p += count * 1;
        }
    }
#endif
    return p - (char *) ptr;
}

    size_t
fwrite_native(void *ptr, size_t size, size_t nmemb, FILE * stream)
{
    size_t r;
#ifndef IS_LITTLE_ENDIAN
    r = fwrite(ptr, size, nmemb, stream);
#else
# define BLOCK 4096
    size_t	nelem, n, ret;
    char	buffer[BLOCK];

    switch (size) {
    case 1:
	r = fwrite(ptr, size, nmemb, stream);
	break;
    case 2:
	nelem = nmemb;
	r = 0;
	while (nelem > 0) {
		n = BLOCK / 2;
		if (nelem < n) n = nelem;
		memcpy(buffer, ptr, n * 2);
		SWAB2(buffer, n);
		ret = fwrite(buffer, 2, n, stream);
		if (ret < n) break;
		ptr = (char *)ptr + n * 2;
		nelem -= n;
		r += ret;
	}
	break;
    case 4:
	nelem = nmemb;
	r = 0;
	while (nelem > 0) {
		n = BLOCK / 4;
		if (nelem < n) n = nelem;
		memcpy(buffer, ptr, n * 4);
		SWAB4(buffer, n);
		ret = fwrite(buffer, 4, n, stream);
		if (ret < n) break;
		ptr = (char *)ptr + n * 4;
		nelem -= n;
		r += ret;
	}
	break;
    case 8:
	nelem = nmemb;
	r = 0;
	while (nelem > 0) {
		n = BLOCK / 8;
		if (nelem < n) n = nelem;
		memcpy(buffer, ptr, n * 8);
		SWAB8(buffer, n);
		ret = fwrite(buffer, 8, n, stream);
		if (ret < n) break;
		ptr = (char *)ptr + n * 8;
		nelem -= n;
		r += ret;
	}
	break;
    case 16:
	nelem = nmemb;
	r = 0;
	while (nelem > 0) {
		n = BLOCK / 16;
		if (nelem < n) n = nelem;
		memcpy(buffer, ptr, n * 16);
		SWAB16(buffer, n);
		ret = fwrite(buffer, 16, n, stream);
		if (ret < n) break;
		ptr = (char *)ptr + n * 16;
		nelem -= n;
		r += ret;
	}
	break;
    default:
	fprintf(stderr,  "fwrite_native(%p, invalid:%lu, %lu, %p)\n",
		ptr, (unsigned long)size, (unsigned long)nmemb, stream);
	return -1;
    }
#endif
    return r;
}

    size_t
fwrite_fmt(void *ptr, char *fmt, FILE * stream)
{
    char *p, *f, *f2;
    int conv;
    unsigned long count;
    size_t sum;
#ifdef IS_LITTLE_ENDIAN
    char *buffer, *tail;
    int dsize, ii, jj;
#endif

    sum = 0;
    p = (char*)ptr;
    for (f = fmt; *f; f++) {
	count = strtoul(f, &f2, 0);
	if (f2 == f) {
	    count = 1;
	    conv = *f2;
	} else if (f2 == NULL) {
	    conv = 'b';
	    /* this causes loop-break after f++ */
	    f = "*";
	} else {
	    conv = *f2;
	    f = f2;
	}
        if(conv == 'q' || conv == 'Q'){
            sum += 16 * count;
        } else if (conv == 'd' || conv == 'D' || conv == 'l'
                 || conv == 'L'){
            sum += 8 * count;
	} else if (conv == 'f' || conv == 'F' || conv == 'i'
		   || conv == 'I') {
            sum += 4 * count;
	} else if (conv == 'h' || conv == 'H') {
            sum +=  2 * count;
	} else {
            sum += 1 * count;
        }            

    }
#ifndef IS_LITTLE_ENDIAN
    fwrite(p, sum, 1, stream);
    p += sum;
#else    
    if((buffer = (char *)malloc(sum)) == NULL){
        fprintf(stderr, "malloc error, %s, %d\n", __FILE__, __LINE__ - 1);
        return 0;
    }

    tail = buffer;
    for (f = fmt; *f; f++) {
	count = strtoul(f, &f2, 0);
	if (f2 == f) {
	    count = 1;
	    conv = *f2;
	} else if (f2 == NULL) {
	    conv = 'b';
	    /* this causes loop-break after f++ */
	    f = "*";
	} else {
	    conv = *f2;
	    f = f2;
	}
	if (conv == 'q' || conv == 'Q') {
            dsize = 16;
	} else if (conv == 'd' || conv == 'D' || conv == 'l'
		   || conv == 'L') {
            dsize = 8;
	} else if (conv == 'f' || conv == 'F' || conv == 'i'
		   || conv == 'I') {
            dsize = 4;
	} else if (conv == 'h' || conv == 'H') {
            dsize = 2;
	} else {
            dsize = 1;
	}
        for(jj = 0; jj < count; jj++){
            for(ii = 0; ii < dsize; ii++){
                tail[dsize * jj + ii] = p[dsize * jj + dsize - (ii + 1)];
            }
        }
        tail += dsize * count;
        p += dsize * count;
    }
    fwrite(buffer, sum, 1, stream);
    free(buffer);
#endif
    return p - (char *) ptr;
}


/*
 * ^ϊ[`
 */

/*
 * big endian z 4 oCg long l
 */
	long
btol(const unsigned char *ptr)
{
	long	r;
	r = ptr[0];
	r <<= 8;
	r |= ptr[1];
	r <<= 8;
	r |= ptr[2];
	r <<= 8;
	r |= ptr[3];
	return r;
}

/*
2003.2.24 tabito added
byte order ϊ[`
*/

void 
endian_swab2(void *ptr, size_t nmemb)
{
#ifdef IS_LITTLE_ENDIAN
    char *cptr;

    cptr = (char*)ptr;
    SWAB2(cptr,nmemb);
#endif
    return;
}

void 
endian_swab4(void *ptr, size_t nmemb)
{
#ifdef IS_LITTLE_ENDIAN
    char *cptr;

    cptr = (char*)ptr;
    SWAB4(cptr,nmemb);
#endif
    return;
}

void 
endian_swab8(void *ptr, size_t nmemb)
{
#ifdef IS_LITTLE_ENDIAN
    char *cptr;

    cptr = (char*)ptr;
    SWAB8(cptr,nmemb);
#endif
    return;
}

void 
endian_swab16(void *ptr, size_t nmemb)
{
#ifdef IS_LITTLE_ENDIAN
    char *cptr;

    cptr = (char*)ptr;
    SWAB16(cptr,nmemb);
#endif
    return;
}

void endian_swab(void *ptr, size_t size, size_t nmemb)
{
#ifdef IS_LITTLE_ENDIAN
    switch(size){
    case 2:
        endian_swab2(ptr, nmemb);
        break;
    case 4:
        endian_swab4(ptr, nmemb);
        break;
    case 8:
        endian_swab8(ptr, nmemb);
        break;
    case 16:
        endian_swab16(ptr, nmemb);
        break;
    default:
        break;

    }
#endif
    return;

}

char* nus_memcpy_native_w(void *dst, const void *src, size_t size, size_t nmemb)
{
    char *dst_w, *src_w;
#ifdef IS_LITTLE_ENDIAN
    int ii, jj;
#endif

    dst_w = (char*) dst;
    src_w = (char*) src;
#ifndef IS_LITTLE_ENDIAN    
    memcpy(dst_w, src_w, size * nmemb);
#else
    for(jj = 0; jj < nmemb; jj++){
        for(ii = 0; ii < size; ii++){
            dst_w[size * jj + ii] = src_w[size * jj + size - (ii + 1)];
        }
    }

#endif
    return dst_w + size * nmemb;

}

char* nus_memcpy_native_r(void *dst, const void *src, size_t size, size_t nmemb)
{
    char *dst_w, *src_w;
#ifdef IS_LITTLE_ENDIAN
    int ii, jj;
#endif

    dst_w = (char*) dst;
    src_w = (char*) src;
#ifndef IS_LITTLE_ENDIAN    
    memcpy(dst_w, src_w, size * nmemb);
#else
    for(jj = 0; jj < nmemb; jj++){
        for(ii = 0; ii < size; ii++){
            dst_w[size * jj + ii] = src_w[size * jj + size - (ii + 1)];
        }
    }

#endif
    return src_w + size * nmemb;

}

size_t
nus_memcpy_fmt(void *ptr, void *src, char *fmt)
{
    char *p, *f, *f2;
    int conv;
    unsigned long count;
    size_t sum;
    char *tail;
#ifdef IS_LITTLE_ENDIAN
    int ii, jj;
    size_t dsize;
#endif

    sum = 0;
    for (f = fmt; *f; f++) {
	count = strtoul(f, &f2, 0);
	if (f2 == f) {
	    count = 1;
	    conv = *f2;
	} else if (f2 == NULL) {
	    conv = 'b';
	    /* this causes loop-break after f++ */
	    f = "*";
	} else {
	    conv = *f2;
	    f = f2;
	}
        if(conv == 'q' || conv == 'Q'){
            sum += 16 * count;
        } else if (conv == 'd' || conv == 'D' || conv == 'l'
                 || conv == 'L'){
            sum += 8 * count;
	} else if (conv == 'f' || conv == 'F' || conv == 'i'
		   || conv == 'I') {
            sum += 4 * count;
	} else if (conv == 'h' || conv == 'H') {
            sum +=  2 * count;
	} else {
            sum += 1 * count;
        }            

    }
    p = (char*)ptr;
    tail = (char*)src;
#ifndef IS_LITTLE_ENDIAN
    memcpy(p, tail, sum);
    p += sum;
#else    
    for (f = fmt; *f; f++) {
	count = strtoul(f, &f2, 0);
	if (f2 == f) {
	    count = 1;
	    conv = *f2;
	} else if (f2 == NULL) {
	    conv = 'b';
	    /* this causes loop-break after f++ */
	    f = "*";
	} else {
	    conv = *f2;
	    f = f2;
	}
	if (conv == 'q' || conv == 'Q') {
            dsize = 16;
	} else if (conv == 'd' || conv == 'D' || conv == 'l'
		   || conv == 'L') {
            dsize = 8;
	} else if (conv == 'f' || conv == 'F' || conv == 'i'
		   || conv == 'I') {
            dsize = 4;
	} else if (conv == 'h' || conv == 'H') {
            dsize = 2;
	} else {
            dsize = 1;
	}
        for(jj = 0; jj < count; jj++){
            for(ii = 0; ii < dsize; ii++){
                p[dsize * jj + ii] = tail[dsize * jj + dsize - (ii + 1)];
            }
        }
        tail += dsize * count;
        p += dsize * count;
    }
#endif
    return p - (char*)ptr;
}

void endian_swab_fmt(void *ptr, char* fmt)
{
#ifdef IS_LITTLE_ENDIAN
    char *p, *f, *f2;
    int conv;
    unsigned long count;
    size_t dsize;

    p = (char*)ptr;
    for (f = fmt; *f; f++) {
	count = strtoul(f, &f2, 0);
	if (f2 == f) {
	    count = 1;
	    conv = *f2;
	} else if (f2 == NULL) {
	    conv = 'b';
	    /* this causes loop-break after f++ */
	    f = "*";
	} else {
	    conv = *f2;
	    f = f2;
	}
	if (conv == 'q' || conv == 'Q') {
            dsize = 16;
	} else if (conv == 'd' || conv == 'D' || conv == 'l'
		   || conv == 'L') {
            dsize = 8;
	} else if (conv == 'f' || conv == 'F' || conv == 'i'
		   || conv == 'I') {
            dsize = 4;
	} else if (conv == 'h' || conv == 'H') {
            dsize = 2;
	} else {
            dsize = 1;
	}
        endian_swab(p, dsize, count);
        p += dsize * count;
    }
#endif
    return;

}

