#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "typedef.h"
#include "grib2_io.h"
#include "endian.h"
#include <math.h>
#include <float.h>

#undef DEBUG

#define BUF_SIZE 1024

static void grib2_read_sec3_ln(N_UI1* buf, sec3_proj_30_t *sec3_ln);
static void grib2_read_sec3_ll(N_UI1* buf, sec3_proj_0_t *sec3_ll);
static void grib2_read_sec4_prod_0(N_UI1 buf[], sec4_prod_0_t *sec4_prod_0);
static void grib2_read_sec4_prod_8(N_UI1 buf[], N_UI2 template, 
                                  sec4_prod_8_t *sec4_prod_8);
static void grib2_read_sec5_rep_0(N_UI1 *buf, sec5_rep_0_t *sec5_rep_0);
static void grib2_read_sec5_rep_200(N_UI1 *buf, sec5_rep_200_t *sec5_rep_200);

static int grib2_decode_rep_0(FILE *fp, grib2_data_list_t *data_list, float *odata){

  N_UI4 nxy;
  N_UI4 n_point;
  unsigned char *buf;
  unsigned int *dout;
  sec5_rep_0_t *rep;
  N_UI4 i;
  N_UI4 j;
  double unit2;
  double unit10;
  N_UI4 length;
  N_UI4 length_bm;

  n_point = data_list->sec5->n_point;

  if ((dout = (unsigned int*)malloc(n_point * sizeof(unsigned int))) == NULL){
    fprintf(stderr, "malloc error at line %d in %s.\n", __LINE__, __FILE__);
    return ERR_MALLOC;
  }


  length = data_list->sec7->length - 5; /* 5: size of header */
  if ((buf = (unsigned char*)malloc(length)) == NULL){
    fprintf(stderr, "malloc error at line %d in %s: size=%d.\n", 
            __LINE__, __FILE__, length);
    return ERR_MALLOC;
  }

  fseek(fp, data_list->sec7->value_pos, SEEK_SET);

  fread(buf, 1, length, fp);

  rep = (sec5_rep_0_t*)(data_list->sec5->rep);

  n_nbit_unpack(buf, length, dout,
                n_point, rep->n_bit);

  free(buf);

  unit2 = pow(2.0, rep->binary_unit);
  unit10 = pow(10.0, -rep->decimal_unit);


  nxy = data_list->sec3->n_grid;
  if (data_list->sec6->bitmap_flag == 0){ /* use bitmap */
    length_bm = data_list->sec6->length - 6;
    if (length_bm < (nxy - 1) / 8 + 1){
      fprintf(stderr, "length_bm is insufficient\n");
      return ERR_GRIB2_SHORT_BUF;
    }
    if ((buf = (unsigned char*)malloc(length_bm)) == NULL){
      fprintf(stderr, "malloc error at line %d in %s: size=%d.\n", 
	      __LINE__, __FILE__, length);
      return ERR_MALLOC;
    }
    fseek(fp, data_list->sec6->bitmap_pos, SEEK_SET);    
    fread(buf, 1, length_bm, fp);
    j = 0;
    for (i = 0; i < nxy; i++){
      if ((buf[i / 8] & (1 << (7 - i % 8))) != 0){
	if (j >= n_point){
	  fprintf(stderr, "the number of packed points is insufficient\n");
	  free(buf);
	  return ERR_GRIB2_SHORT_BUF;
	}
	odata[i] = (rep->reference + dout[j++] * unit2) * unit10;	
      } else{
	odata[i] = MISSING;
      }
    }
    free(buf);
  } else { /* not use bit map: nxy = n_point */
    for (i = 0; i < nxy; i++){
      odata[i] = (rep->reference + dout[i] * unit2) * unit10;
    }
  }



  free(dout);
  return 0;
}

static int grib2_decode_rep_200(FILE *fp, grib2_data_list_t *data_list, float *odata){

  int nxy;
  int nx;
  int ny;
  unsigned char *buf;
  unsigned int *dout;
  sec5_rep_200_t *rep;
  int i;
  double unit10;
  unsigned int length;

  switch(data_list->sec3->grid_template){
  case(30):
    nx = ((sec3_proj_30_t*)(data_list->sec3->proj))->nx;
    ny = ((sec3_proj_30_t*)(data_list->sec3->proj))->ny;
    break;
  case(0):
    nx = ((sec3_proj_0_t*)(data_list->sec3->proj))->nx;
    ny = ((sec3_proj_0_t*)(data_list->sec3->proj))->ny;
    break;
  }

  nxy = nx * ny;
  if ((dout = (unsigned int*)malloc(nxy * sizeof(unsigned int))) == NULL){
    fprintf(stderr, "malloc error at line %d in %s.\n", __LINE__, __FILE__);
    return ERR_MALLOC;
  }

  length = data_list->sec7->length;
  if ((buf = (unsigned char*)malloc(length)) == NULL){
    fprintf(stderr, "malloc error at line %d in %s: size=%d.\n", 
            __LINE__, __FILE__, length);
    return ERR_MALLOC;
  }

  fseek(fp, data_list->sec7->value_pos, SEEK_SET);

  fread(buf, 1, length, fp);

  rep = (sec5_rep_200_t*)(data_list->sec5->rep);

  unit10 = pow(10.0, -rep->unit);
  n_decode_rlen_nbit((N_UI1*)dout, buf, length, nxy,
		     rep->max_level_this, rep->n_bit);

  for (i = 0; i < nxy; i++){
    if (dout[i] == 0) {
      odata[i] = MISSING;
    } else{
      odata[i] = rep->lev_to_prec[dout[i] - 1] * unit10;
    }
  }
  
  free(buf);
  free(dout);
  return 0;
}

static grib2_data_list_t* 
lookup(const grib2_data_t *data, const ymdhm_t *ymdhm, const N_SI4 term, 
       const N_UI1 param_category, const N_UI1 param_number, 
       const N_UI1 layer_kind, const double layer){
  
  int term_unit;
  int term_data;
  double layer_unit;
  double layer_data;
  grib2_data_list_t *found;
  grib2_data_list_t *p;
  sec4_prod_0_t *prod;
  int i;
  N_UI2 year = ymdhm->year;
  N_UI1 month = ymdhm->month;
  N_UI1 day = ymdhm->day;
  N_UI1 hour = ymdhm->hour;
  N_UI1 min = ymdhm->min;

  found = NULL;
  for (p = data->list; p != NULL; p = p->next){

    prod = (sec4_prod_0_t*)(p->sec4->prod_def);
    if (prod == NULL){
      fprintf(stderr, "There is an undefined prod_def.\n");
      continue;
    }

    if (data->list->sec1->year == year &&
	data->list->sec1->month == month &&
	data->list->sec1->day == day &&
	data->list->sec1->hour == hour &&
	data->list->sec1->min == min){
      if (prod->term_unit == 0) {
	term_unit = 1;
      } else if (prod->term_unit == 1) {
	term_unit = 60;
      } else {
	fprintf(stderr, "Not supported term unit.\n");
	return NULL;
      }
      term_data = prod->term * term_unit;
      if (term_data == term && 
	  prod->param_category == param_category &&
	  prod->param_number == param_number){
	if (prod->fixed_layer1_unit != 0xff){
	  layer_unit = 1.0;
	    for (i = 0; i < prod->fixed_layer1_unit; i++){
	      layer_unit *= 0.1;
	    }
	  layer_data = prod->fixed_layer1 * layer_unit;
	  if (fabs(layer_data - layer) < 1.0e-6 && 
	      prod->fixed_layer1_kind == layer_kind){
	    found = p;
	    break;
	  }
	} else if (prod->fixed_layer1_kind == layer_kind) {
	  found = p;
	  break;
	}
      }
    }
  }
  return found;
}





static int add_rec_list(unsigned int length, unsigned char section, 
			void *rec_each, grib2_rec_t *rec){
  grib2_rec_list_t *add_elem;
  
  if ((add_elem = (grib2_rec_list_t*)malloc(sizeof(grib2_data_list_t))) == NULL){
    fprintf(stderr, "malloc error at line %d in %s.\n", __LINE__, __FILE__);
    return ERR_MALLOC;
  }

  add_elem->section = section;
  add_elem->length = length;
  add_elem->rec = rec_each;
  add_elem->next = NULL;

  if (rec->list == NULL){
    rec->list = add_elem;
    rec->last = add_elem;

  } else {
    rec->last->next = add_elem;
    rec->last = add_elem;
  }
  return 0;
}

static int add_data_list(sec0_t *sec0, sec1_t *sec1, 
			 sec2_t *sec2, sec3_t *sec3, sec4_t *sec4, 
			 sec5_t *sec5, sec6_t *sec6, sec7_t *sec7, 
			 grib2_data_t *data){
  grib2_data_list_t *add_elem;
  
  if ((add_elem = (grib2_data_list_t*)malloc(sizeof(grib2_data_list_t))) == NULL){
    fprintf(stderr, "malloc error at line %d in %s.\n", __LINE__, __FILE__);
    return ERR_MALLOC;
  }

  add_elem->sec0 = sec0;
  add_elem->sec1 = sec1;
  add_elem->sec2 = sec2;
  add_elem->sec3 = sec3;
  add_elem->sec4 = sec4;
  add_elem->sec5 = sec5;
  add_elem->sec6 = sec6;
  add_elem->sec7 = sec7;

  add_elem->next = NULL;

  if (data->list == NULL){
    data->list = add_elem;
    data->last = add_elem;

  } else {
    data->last->next = add_elem;
    data->last = add_elem;
  }
  return 0;
}

static void arrange_info(grib2_rec_t *rec, grib2_data_t *data){

  grib2_rec_list_t *p;
  sec0_t *cur_sec0 = NULL;
  sec1_t *cur_sec1 = NULL;
  sec2_t *cur_sec2 = NULL;
  sec3_t *cur_sec3 = NULL;
  sec4_t *cur_sec4 = NULL;
  sec5_t *cur_sec5 = NULL;
  sec6_t *cur_sec6 = NULL;
  sec7_t *cur_sec7 = NULL;

  for (p = rec->list; p != NULL; p = p->next){
    if (p->section == 0){
      cur_sec0 = (sec0_t*)(p->rec);
    } else if (p->section == 1){
      cur_sec1 = (sec1_t*)(p->rec);
    } else if (p->section == 2){
      cur_sec2 = (sec2_t*)(p->rec);
    } else if (p->section == 3){
      cur_sec3 = (sec3_t*)(p->rec);
    } else if (p->section == 4){
      cur_sec4 = (sec4_t*)(p->rec);
    } else if (p->section == 5){
      cur_sec5 = (sec5_t*)(p->rec);
    } else if (p->section == 6){
      cur_sec6 = (sec6_t*)(p->rec);
    } else if (p->section == 7){
      cur_sec7 = (sec7_t*)(p->rec);
      add_data_list(cur_sec0, cur_sec1, 
		    cur_sec2, cur_sec3, cur_sec4, cur_sec5, cur_sec6, cur_sec7, 
		    data);
    }
  }
  return;
}



static void query_section(FILE *fp, N_UI4 *length, N_UI1 *sec){
    N_UI1 buf[BUF_SIZE];

    fread(buf, 4, 1, fp);

    if (memcmp(buf, "GRIB", 4) == 0){
      *length = 16;
      *sec = 0;
      fseek(fp, -4L, SEEK_CUR);
    } else if (memcmp(buf, "7777", 4) == 0){
      *length = 4;
      *sec = 8;
      fseek(fp, -4L, SEEK_CUR);
    } else {
      fread(buf + 4, 1, 1, fp);
      memcpy(length, buf, 4);
      endian_swab4(length, 1);
      *sec = buf[4];
      fseek(fp, -5L, SEEK_CUR);
    }
    return;
}

static int grib2_read_sec0(FILE *fp, sec0_t *sec0){
    N_UI1 buf[BUF_SIZE];
    size_t size;
    
    size = fread(buf, 1, 16, fp);
    if (size != 16){
        fprintf(stderr, "Read Error, sec0. at line %d in %s.\n", 
		__LINE__, __FILE__);
        return ERR_GRIB2_SEC0;
    }
    if (memcmp(buf, "GRIB", 4) != 0){
        fprintf(stderr, "Not a GRIB file at line %d in %s.\n", 
		__LINE__, __FILE__);
        return ERR_GRIB2_SEC0_NO_GRIB;
    }
    memcpy(&sec0->master, buf + 6, 1);
    memcpy(&sec0->version, buf + 7, 1);
    if (sec0->version != 2){
      fprintf(stderr, "A version number is not 2. at line %d in %s.\n", 
	      __LINE__, __FILE__);
        return ERR_GRIB2_SEC0_NO_GRIB2;
    }
    memcpy(&sec0->length, buf + 8, 8);
    endian_swab8(&(sec0->length), 1);

#ifdef DEBUG
    grib2_print_sec0(sec0);
#endif
    return 0;
}


static int grib2_read_sec1(FILE *fp, sec1_t *sec1){
    N_UI1 buf[BUF_SIZE];
    size_t size;
    N_UI4 length;
    
    size = fread(&length, 1, 4, fp);
    endian_swab4(&length, 1);

    fseek(fp, -4, SEEK_CUR);

    size = fread(buf, 1, length, fp);    
    if (size != length){
        fprintf(stderr, "Read Error, sec1. at line %d in %s.\n", 
		__LINE__, __FILE__);
        return ERR_GRIB2_SEC1;
    }
    memcpy(&sec1->length, buf, 4);
    endian_swab4(&sec1->length, 1);

    sec1->section = buf[4];
    
    memcpy(&sec1->center, buf + 5, 2);
    endian_swab2(&sec1->center, 1);

    memcpy(&sec1->subcenter, buf + 7, 2);
    endian_swab2(&sec1->subcenter, 1);
    
    sec1->table_version = buf[9];
    sec1->region_version = buf[10];
    sec1->vtime_mean = buf[11];

    memcpy(&sec1->year, buf + 12, 2);
    endian_swab2(&sec1->year, 1);

    sec1->month = buf[14];
    sec1->day = buf[15];
    sec1->hour = buf[16];
    sec1->min = buf[17];
    sec1->sec = buf[18];
    sec1->status = buf[19];
    sec1->kind = buf[20];

#ifdef DEBUG
    grib2_print_sec1(sec1);
#endif

    return 0;
}



static int grib2_read_sec2(FILE *fp, sec1_t *sec2){
    N_UI1 buf[BUF_SIZE];
    size_t size;
    N_UI4 length;
 
    size = fread(&length, 1, 4, fp);
    endian_swab4(&length, 1);

    fseek(fp, -4, SEEK_CUR);

    size = fread(buf, 1, length, fp);    
    if (size != length){
        fprintf(stderr, "Read Error, sec2. at line %d in %s.\n", 
		__LINE__, __FILE__);
        return ERR_GRIB2_SEC2;
    }
    memcpy(&sec2->length, buf, 4);
    endian_swab4(&sec2->length, 1);

    sec2->section = buf[4];


#ifdef DEBUG
    grib2_print_sec2(sec2);
#endif

    return 0;
}



static int grib2_read_sec3(FILE *fp, sec3_t *sec3){
    N_UI1 buf[BUF_SIZE];
    size_t size;
    N_UI4 length;

    size = fread(&length, 1, 4, fp);
    endian_swab4(&length, 1);

    fseek(fp, -4, SEEK_CUR);

    size = fread(buf, 1, length, fp);    
    if (size != length){
        fprintf(stderr, "Read Error, sec3. at line %d in %s.\n",
		__LINE__, __FILE__);
        return ERR_GRIB2_SEC3;
    }
    memcpy(&sec3->length, buf, 4);
    endian_swab4(&sec3->length, 1);

    sec3->section = buf[4];
    sec3->source = buf[5];

    memcpy(&sec3->n_grid, buf + 6, 4);
    endian_swab4(&sec3->n_grid, 1);

    sec3->oct_optlist = buf[10];
    sec3->list_kind = buf[11];
    memcpy(&sec3->grid_template, buf + 12, 2);
    endian_swab2(&sec3->grid_template, 1);

    if (sec3->grid_template == 30){
        sec3->proj = (sec3_proj_30_t*)malloc(sizeof(sec3_proj_30_t));
        grib2_read_sec3_ln(buf, sec3->proj);
    } else if (sec3->grid_template == 0){
        sec3->proj = (sec3_proj_0_t*)malloc(sizeof(sec3_proj_0_t));
        grib2_read_sec3_ll(buf, sec3->proj);
    }

#ifdef DEBUG
    grib2_print_sec3(sec3);
#endif


    return 0;
}


static void grib2_read_sec3_ln(N_UI1* buf, sec3_proj_30_t *sec3_ln){
    sec3_ln->earth_shape = buf[14];

    sec3_ln->r_earth_unit = buf[15];
    memcpy(&sec3_ln->r_earth, buf + 16, 4);
    endian_swab4(&sec3_ln->r_earth, 1);

    sec3_ln->a_earth_unit = buf[20];
    memcpy(&sec3_ln->a_earth, buf + 21, 4);
    endian_swab4(&sec3_ln->a_earth, 1);

    sec3_ln->b_earth_unit = buf[25];
    memcpy(&sec3_ln->b_earth, buf + 26, 4);
    endian_swab4(&sec3_ln->b_earth, 1);

    memcpy(&sec3_ln->nx, buf + 30, 4);
    endian_swab4(&sec3_ln->nx, 1);

    memcpy(&sec3_ln->ny, buf + 34, 4);
    endian_swab4(&sec3_ln->ny, 1);

    memcpy(&sec3_ln->la1, buf + 38, 4);
    endian_swab4(&sec3_ln->la1, 1);

    memcpy(&sec3_ln->lo1, buf + 42, 4);
    endian_swab4(&sec3_ln->lo1, 1);

    sec3_ln->res = buf[46];    

    memcpy(&sec3_ln->lad, buf + 47, 4);
    endian_swab4(&sec3_ln->lad, 1);

    memcpy(&sec3_ln->lov, buf + 51, 4);
    endian_swab4(&sec3_ln->lov, 1);

    memcpy(&sec3_ln->dx, buf + 55, 4);
    endian_swab4(&sec3_ln->dx, 1);

    memcpy(&sec3_ln->dy, buf + 59, 4);
    endian_swab4(&sec3_ln->dy, 1);

    sec3_ln->flag_center = buf[63];
    sec3_ln->scan_mode = buf[64];

    memcpy(&sec3_ln->latin1, buf + 65, 4);
    endian_swab4(&sec3_ln->latin1, 1);

    memcpy(&sec3_ln->latin2, buf + 69, 4);
    endian_swab4(&sec3_ln->latin2, 1);
    
    memcpy(&sec3_ln->proj_sp_lat, buf + 73, 4);
    endian_swab4(&sec3_ln->proj_sp_lat, 1);

    memcpy(&sec3_ln->proj_sp_lon, buf + 77, 4);
    endian_swab4(&sec3_ln->proj_sp_lon, 1);

    return;
}

static void grib2_read_sec3_ll(N_UI1* buf, sec3_proj_0_t *sec3_ll){
    sec3_ll->earth_shape = buf[14];

    sec3_ll->r_earth_unit = buf[15];
    memcpy(&sec3_ll->r_earth, buf + 16, 4);
    endian_swab4(&sec3_ll->r_earth, 1);

    sec3_ll->a_earth_unit = buf[20];
    memcpy(&sec3_ll->a_earth, buf + 21, 4);
    endian_swab4(&sec3_ll->a_earth, 1);

    sec3_ll->b_earth_unit = buf[25];
    memcpy(&sec3_ll->b_earth, buf + 26, 4);
    endian_swab4(&sec3_ll->b_earth, 1);

    memcpy(&sec3_ll->nx, buf + 30, 4);
    endian_swab4(&sec3_ll->nx, 1);

    memcpy(&sec3_ll->ny, buf + 34, 4);
    endian_swab4(&sec3_ll->ny, 1);

    memcpy(&sec3_ll->basic_angle, buf + 38, 4);
    endian_swab4(&sec3_ll->basic_angle, 1);

    memcpy(&sec3_ll->subdivision, buf + 42, 4);
    endian_swab4(&sec3_ll->basic_angle, 1);

    memcpy(&sec3_ll->la1, buf + 46, 4);
    endian_swab4(&sec3_ll->la1, 1);

    memcpy(&sec3_ll->lo1, buf + 50, 4);
    endian_swab4(&sec3_ll->lo1, 1);

    sec3_ll->res = buf[54];    

    memcpy(&sec3_ll->la2, buf + 55, 4);
    endian_swab4(&sec3_ll->la2, 1);

    memcpy(&sec3_ll->lo2, buf + 59, 4);
    endian_swab4(&sec3_ll->lo2, 1);

    memcpy(&sec3_ll->dx, buf + 63, 4);
    endian_swab4(&sec3_ll->dx, 1);

    memcpy(&sec3_ll->dy, buf + 67, 4);
    endian_swab4(&sec3_ll->dy, 1);

    sec3_ll->scan_mode = buf[71];
    return;

}


static int grib2_read_sec4(FILE *fp, sec4_t *sec4){
    N_UI1 buf[BUF_SIZE];
    size_t size;
    N_UI4 length;

    size = fread(&length, 1, 4, fp);
    endian_swab4(&length, 1);

    fseek(fp, -4, SEEK_CUR);

    size = fread(buf, 1, length, fp);    
    if (size != length){
        fprintf(stderr, "Read Error, sec4. at line %d in %s.\n",
		__LINE__, __FILE__);
        return ERR_GRIB2_SEC4;
    }
    memcpy(&sec4->length, buf, 4);
    endian_swab4(&sec4->length, 1);
    
    sec4->section = buf[4];

    memcpy(&sec4->n_grid_after_temp, buf + 5, 2);
    endian_swab2(&sec4->n_grid_after_temp, 1);

    memcpy(&sec4->prod_template, buf + 7, 2);
    endian_swab2(&sec4->prod_template, 1);
    
    switch (sec4->prod_template){
    case(0):
      sec4->prod_def = (sec4_prod_0_t*)malloc(sizeof(sec4_prod_0_t));
      grib2_read_sec4_prod_0(buf, sec4->prod_def);
      break;
    case (8):
    case (50008):
      sec4->prod_def = (sec4_prod_8_t*)malloc(sizeof(sec4_prod_8_t));
      grib2_read_sec4_prod_8(buf, sec4->prod_template, sec4->prod_def);
      break;
    }

#ifdef DEBUG
    grib2_print_sec4(sec4);
#endif

    return 0;
}



static void grib2_read_sec4_prod_0(N_UI1 buf[], sec4_prod_0_t *sec4_prod_0){
    
    sec4_prod_0->param_category = buf[9];
    sec4_prod_0->param_number = buf[10];
    sec4_prod_0->kind_process = buf[11];
    sec4_prod_0->bg_process = buf[12];
    sec4_prod_0->process = buf[13];

    memcpy(&sec4_prod_0->cutoff_time_hour, buf + 14, 2);
    endian_swab2(&sec4_prod_0->cutoff_time_hour, 1);

    sec4_prod_0->cutoff_time_min = buf[16];
    sec4_prod_0->term_unit = buf[17];

    memcpy(&sec4_prod_0->term, buf + 18, 4);
    endian_swab4(&sec4_prod_0->term, 1);
    if ((*((N_UI4*)&sec4_prod_0->term) & ((N_UI4)1<<31)) != 0){
      sec4_prod_0->term = -((*((N_UI4*)&sec4_prod_0->term)) & 0x7fffffff);
    }

    sec4_prod_0->fixed_layer1_kind = buf[22];
    sec4_prod_0->fixed_layer1_unit = buf[23];

    memcpy(&sec4_prod_0->fixed_layer1, buf + 24, 4);
    endian_swab4(&sec4_prod_0->fixed_layer1, 1);

    sec4_prod_0->fixed_layer2_kind = buf[28];
    sec4_prod_0->fixed_layer2_unit = buf[29];

    memcpy(&sec4_prod_0->fixed_layer2, buf + 30, 4);
    endian_swab4(&sec4_prod_0->fixed_layer2, 1);

    return;
}

static void grib2_read_sec4_prod_8(N_UI1 buf[], N_UI2 template, 
			   sec4_prod_8_t *sec4_prod_8){
  int i;
    
    sec4_prod_8->param_category = buf[9];
    sec4_prod_8->param_number = buf[10];
    sec4_prod_8->kind_process = buf[11];
    sec4_prod_8->bg_process = buf[12];
    sec4_prod_8->process = buf[13];

    memcpy(&sec4_prod_8->cutoff_time_hour, buf + 14, 2);
    endian_swab2(&sec4_prod_8->cutoff_time_hour, 1);

    sec4_prod_8->cutoff_time_min = buf[16];
    sec4_prod_8->term_unit = buf[17];

    memcpy(&sec4_prod_8->term, buf + 18, 4);
    endian_swab4(&sec4_prod_8->term, 1);
    if ((*((N_UI4*)&sec4_prod_8->term) & ((N_UI4)1<<31)) != 0){
      sec4_prod_8->term = -((*((N_UI4*)&sec4_prod_8->term)) & 0x7fffffff);
    }

    sec4_prod_8->fixed_layer1_kind = buf[22];
    sec4_prod_8->fixed_layer1_unit = buf[23];

    memcpy(&sec4_prod_8->fixed_layer1, buf + 24, 4);
    endian_swab4(&sec4_prod_8->fixed_layer1, 1);

    sec4_prod_8->fixed_layer2_kind = buf[28];
    sec4_prod_8->fixed_layer2_unit = buf[29];

    memcpy(&sec4_prod_8->fixed_layer2, buf + 30, 4);
    endian_swab4(&sec4_prod_8->fixed_layer2, 1);

    memcpy(&sec4_prod_8->eyear, buf + 34, 2);
    endian_swab2(&sec4_prod_8->eyear, 1);

    sec4_prod_8->emonth = buf[36];
    sec4_prod_8->eday = buf[37];
    sec4_prod_8->ehour = buf[38];
    sec4_prod_8->emin = buf[39];
    sec4_prod_8->esec = buf[40];
    sec4_prod_8->n = buf[41];

    memcpy(&sec4_prod_8->n_miss, buf + 42, 4);
    endian_swab4(&sec4_prod_8->n_miss, 1);

    sec4_prod_8->stat = (sec4_prod_8_stat_t**)malloc(
      sizeof(sec4_prod_8_stat_t*) * sec4_prod_8->n);
    for (i = 0; i < sec4_prod_8->n; i++){
      sec4_prod_8->stat[i] = (sec4_prod_8_stat_t*)malloc(sizeof(sec4_prod_8_stat_t));
      sec4_prod_8->stat[i]->stat_process = buf[46 + 12 * i];
      sec4_prod_8->stat[i]->delt_kind = buf[47 + 12 * i];
      sec4_prod_8->stat[i]->term_unit = buf[48 + 12 * i];

      memcpy(&sec4_prod_8->stat[i]->term, &buf[49 + 12 * i], 4);
      endian_swab4(&sec4_prod_8->stat[i]->term, 1);

      sec4_prod_8->stat[i]->delt_unit = buf[53 + 12 * i];

      memcpy(&sec4_prod_8->stat[i]->delt_unit, buf + 54 + 12 * i, 4);
      endian_swab4(&sec4_prod_8->stat[i]->delt_unit, 1);
    }

    if (template == 50008){
      memcpy(&sec4_prod_8->radar_use1, 
	     buf + 58 + 12 * (sec4_prod_8->n - 1), 8);
      endian_swab8(&sec4_prod_8->radar_use1, 1);

      memcpy(&sec4_prod_8->radar_use2, 
	     buf + 66 + 12 * (sec4_prod_8->n - 1), 8);
      endian_swab8(&sec4_prod_8->radar_use2, 1);

      memcpy(&sec4_prod_8->raingauge_use, 
	     buf + 74 + 12 * (sec4_prod_8->n - 1), 8);
      endian_swab8(&sec4_prod_8->raingauge_use, 1);
    }


    return;
}






static int grib2_read_sec5(FILE *fp, sec5_t *sec5){
    N_UI1 buf[BUF_SIZE];
    size_t size;
    N_UI4 length;

    size = fread(&length, 1, 4, fp);
    endian_swab4(&length, 1);

    fseek(fp, -4, SEEK_CUR);

    size = fread(buf, 1, length, fp);    
    if (size != length){
        fprintf(stderr, "Read Error, sec5. at line %d in %s.\n", 
		__LINE__, __FILE__);
        return ERR_GRIB2_SEC5;
    }
    memcpy(&sec5->length, buf, 4);
    endian_swab4(&sec5->length, 1);
    
    sec5->section = buf[4];
    
    memcpy(&sec5->n_point, buf + 5, 4);
    endian_swab4(&sec5->n_point, 1);
    
    memcpy(&sec5->rep_template, buf + 9, 2);
    endian_swab2(&sec5->rep_template, 1);

    switch (sec5->rep_template){
    case(0):
      sec5->rep = (sec5_rep_0_t*)malloc(sizeof(sec5_rep_0_t));
      grib2_read_sec5_rep_0(buf, sec5->rep);
      break;
    case (200):
      sec5->rep = (sec5_rep_200_t*)malloc(sizeof(sec5_rep_200_t));
      grib2_read_sec5_rep_200(buf, sec5->rep);
      break;
    }

#ifdef DEBUG
    grib2_print_sec5(sec5);
#endif


    return 0;
}

static void grib2_read_sec5_rep_0(N_UI1 *buf, sec5_rep_0_t *sec5_rep_0){
  memcpy(&sec5_rep_0->reference, buf + 11, 4);
  endian_swab4(&sec5_rep_0->reference, 1);

  memcpy(&sec5_rep_0->binary_unit, buf + 15, 2);
  endian_swab2(&sec5_rep_0->binary_unit, 1);
  if (sec5_rep_0->binary_unit < 0){
    sec5_rep_0->binary_unit = (32768 - abs(sec5_rep_0->binary_unit)) * (-1);
  }
  memcpy(&sec5_rep_0->decimal_unit, buf + 17, 2);
  endian_swab2(&sec5_rep_0->decimal_unit, 1);
  if (sec5_rep_0->decimal_unit < 0){
    sec5_rep_0->decimal_unit = (32768 - abs(sec5_rep_0->decimal_unit)) * (-1);
  }
  sec5_rep_0->n_bit = buf[19];
  sec5_rep_0->kind_value = buf[20];

  return;
}

static void grib2_read_sec5_rep_200(N_UI1 *buf, sec5_rep_200_t *sec5_rep_200){
  N_UI2 max_level;

  sec5_rep_200->n_bit = buf[11];

  memcpy(&sec5_rep_200->max_level_this, buf + 12, 2);
  endian_swab2(&sec5_rep_200->max_level_this, 1);

  memcpy(&sec5_rep_200->max_level, buf + 14, 2);
  endian_swab2(&sec5_rep_200->max_level, 1);

  sec5_rep_200->unit = buf[16];

  max_level = sec5_rep_200->max_level;
  sec5_rep_200->lev_to_prec
    = (N_UI2*)malloc(sizeof(N_UI2) * max_level);
  
  memcpy(sec5_rep_200->lev_to_prec, buf + 17, sizeof(N_UI2) * max_level);
  endian_swab2(sec5_rep_200->lev_to_prec, max_level);

  return;
}



static int grib2_read_sec6(FILE *fp, sec6_t *sec6){
  N_UI1 buf[BUF_SIZE];
  N_UI8 pos_s;
  size_t size;

  size = fread(buf, 1, 6, fp);    
  if (size != 6){
    fprintf(stderr, "Read Error, sec6. at line %d in %s.\n",
	    __LINE__, __FILE__);
    return ERR_GRIB2_SEC6;
  }
  memcpy(&sec6->length, buf, 4);
  endian_swab4(&sec6->length, 1);
  
  sec6->section = buf[4];
  sec6->bitmap_flag = buf[5];
  pos_s = ftell(fp);

  if (sec6->bitmap_flag != 0){
    sec6->bitmap_pos = 0xffffffffffffULL;
  } else {
    sec6->bitmap_pos = pos_s;
  }
  fseek(fp, sec6->length - 6, SEEK_CUR);

#ifdef DEBUG
    grib2_print_sec6(sec6);
#endif

  return 0;
}


static int grib2_read_sec7(FILE *fp, sec7_t *sec7){
  N_UI1 buf[BUF_SIZE];
  N_UI8 pos_s;
  size_t size;

  size = fread(buf, 1, 5, fp);    
  if (size != 5){
    fprintf(stderr, "Read Error, sec7. at line %d in %s.\n", 
	    __LINE__, __FILE__);
    return ERR_GRIB2_SEC7;
  }
  memcpy(&sec7->length, buf, 4);
  endian_swab4(&sec7->length, 1);
  
  sec7->section = buf[4];

  pos_s = ftell(fp);
  sec7->value_pos = pos_s;

  fseek(fp, sec7->length - 5, SEEK_CUR);
  

#ifdef DEBUG
    grib2_print_sec7(sec7);
#endif


  return 0;

}

int grib2_inq_nz(grib2_t *grib2, ymdhm_t *ymdhm, N_SI4 term, 
		  N_UI1 parm_cat, N_UI1 parm_num){

  /* term: min */

  int k;
  grib2_data_t *data = &(grib2->data);
  grib2_data_list_t *data_list;

  k = 1;
  while (1){
    data_list = lookup(data, ymdhm, term, 
                       parm_cat, parm_num, 
                       LAYER_HYBRID, (double)k);

    if (data_list == NULL){
      break;
    }
    k++;
  }
  return k - 1;
}

int grib2_read(grib2_t *grib2,
	       const ymdhm_t *ymdhm, const N_SI4 term, 
	       const N_UI1 parm_cat, const N_UI1 parm_num, 
	       const N_UI1 layer_kind, const N_UI4 nz, 
	       const double *layer, float *data_out){

    grib2_data_list_t *data_list;
    grib2_data_t *data = &(grib2->data);
    int kk;
    unsigned int nxy;

    for (kk = 0; kk < nz; kk++){
      data_list = lookup(data, ymdhm, term, 
			 parm_cat, parm_num, 
			 layer_kind, layer[kk]);

      if (data_list == NULL){
	fprintf(stderr, 
		"Not found data: cat = %u, num = %u, layer_kind = %u, layer = %f\n", 
		parm_cat, parm_num, layer_kind, layer[kk]);
	return ERR_GRIB2_NOT_FOUND_DATA;
      }
      nxy = data_list->sec3->n_grid;
      
      switch (data_list->sec5->rep_template){
      case (0):
	grib2_decode_rep_0(grib2->fp, data_list, data_out + nxy * kk);
	break;
      case (200):
	grib2_decode_rep_200(grib2->fp, data_list, data_out + nxy * kk);	
	break;
      }
    }
    return 0;
}


int grib2_read_info(grib2_t *grib2){

  unsigned int length;
  unsigned char sec;
  void *rec_each;
  FILE *fp = grib2->fp;
  grib2_rec_t *rec = &(grib2->rec);
  grib2_data_t *data = &(grib2->data);
  int r;


  while (1){
    query_section(fp, &length, &sec);

    if (sec == 0){
      rec_each = (sec0_t*)malloc(sizeof(sec0_t));
      if ((r = grib2_read_sec0(fp, rec_each))) return r;

    } else if (sec == 1) {
      rec_each = (sec1_t*)malloc(sizeof(sec1_t));
      if ((r = grib2_read_sec1(fp, rec_each))) return r;

    } else if (sec == 2){
      rec_each = (sec2_t*)malloc(sizeof(sec2_t));
      if ((r = grib2_read_sec2(fp, rec_each))) return r;

    } else if (sec == 3){
      rec_each = (sec3_t*)malloc(sizeof(sec3_t));
      if ((r = grib2_read_sec3(fp, rec_each))) return r;

    } else if (sec == 4){
      rec_each = (sec4_t*)malloc(sizeof(sec4_t));
      if ((r = grib2_read_sec4(fp, rec_each))) return r;

    } else if (sec == 5){
      rec_each = (sec5_t*)malloc(sizeof(sec5_t));
      if ((r = grib2_read_sec5(fp, rec_each))) return r;

    } else if (sec == 6){
      rec_each = (sec6_t*)malloc(sizeof(sec6_t));
      if ((r = grib2_read_sec6(fp, rec_each))) return r;

    } else if (sec == 7){
      rec_each = (sec7_t*)malloc(sizeof(sec7_t));
      if ((r = grib2_read_sec7(fp, rec_each))) return r;

    } else if (sec == 8){
      break;
    }
    else {
      fprintf(stderr, "Unknown section number: %u\n", (unsigned int)sec);
      return ERR_GRIB2_UNKNOWN_SEC_NUMBER;
    }
    if ((r = add_rec_list(length, sec, rec_each, rec))) return r;
  }
  arrange_info(rec, data);
  return 0;

}


