#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "typedef.h"
#include "grib2_encode.h"



#undef DEBUG

#define BUF_SIZE 1024

static int grib2_write_sec3_ln(N_UI1* buf, const sec3_proj_30_t *sec3_ln);
static int grib2_write_sec3_ll(N_UI1* buf, const sec3_proj_0_t *sec3_ll);
static int grib2_write_sec4_prod_0(N_UI1 buf[], const sec4_prod_0_t *sec4_prod_0);
static int grib2_write_sec4_prod_8(N_UI1 buf[], const sec4_prod_8_t *sec4_prod_8);
static int grib2_write_sec5_rep_0(N_UI1 *buf, const sec5_rep_0_t *sec5_rep_0);


static int grib2_encode_rep_0(const int nxy, const int n_bit, 
			      const short decimal_unit,
			      const float *fdat, 
			      short *binary_unit, float *reference, 
			      N_UI1 *uc, const int uc_size, 
			      unsigned int *length){
  int ixy;
  int nb;
  int nb2;
  float amp;
  float min =  FLT_MAX; 
  float max = -FLT_MAX; 
  double sf;
  float dat;
  float factor;
  N_UI4 *ui;

  factor = pow(10.0, decimal_unit);

  for (ixy = 0; ixy < nxy; ixy++){
    dat = fdat[ixy] * factor;
    if (dat > max) max = dat;
    if (dat < min) min = dat;
  }
  *reference = min;
  amp = max - min;
  if (amp == 0.0) amp = 1.0;
  frexp((double)amp, &nb);
  *binary_unit = nb - n_bit;
  frexp((double)amp + 0.5 * pow(2.0, *binary_unit), &nb2);
  if (nb != nb2) {
    *binary_unit = *binary_unit + 1;
  }

  sf = dpow2( abs(*binary_unit) );
  if (*binary_unit > 0) sf = 1.0 / sf;
  
  if ((ui = (N_UI4*)malloc(nxy * sizeof(N_UI4))) == NULL){
    fprintf(stderr, "malloc error at line %d in %s.\n", 
            __LINE__, __FILE__);
    return ERR_MALLOC;
  }

  for (ixy = 0; ixy < nxy; ixy++){
    ui[ixy] = (N_UI4)((fdat[ixy] * factor - *reference) * sf + 0.5);
  }
  *length = n_nbit_pack(ui, nxy, uc, uc_size, n_bit);
  free(ui);
  return 0;
}


static int grib2_write_sec0(const grib2_t *grib2){
    N_UI1 buf[BUF_SIZE];
    N_UI8 length;
    sec0_t *sec0 = grib2->data.list->sec0;
    FILE *fp = grib2->fp;

    memcpy(buf, "GRIB", 4);
    buf[4] = 0xff;
    buf[5] = 0xff;

    
    memcpy(buf + 6, &sec0->master, 1);
    memcpy(buf + 7, &sec0->version, 1);

    /* tentative */
    length = 0;
    memcpy(buf + 8, &length,  8);
    endian_swab8(buf + 8, 1);

    fwrite(buf, 1, 16, fp);

#ifdef DEBUG
    printf("=============== written GRIB2 =============\n");
    grib2_print_sec0(sec0);
#endif
    return 0;
}

static int grib2_write_sec1(const grib2_t *grib2){
    N_UI1 buf[BUF_SIZE];
    sec1_t *sec1 = grib2->data.list->sec1;
    FILE *fp = grib2->fp;

    memcpy(buf, &sec1->length, 4);
    endian_swab4(buf, 1);

    buf[4] = sec1->section;
    
    memcpy(buf + 5,  &sec1->center, 2);
    endian_swab2(buf + 5, 1);

    memcpy(buf + 7, &sec1->subcenter, 2);
    endian_swab2(buf + 7, 1);
    
    buf[9] = sec1->table_version;
    buf[10] = sec1->region_version;
    buf[11] = sec1->vtime_mean;

    memcpy(buf + 12, &sec1->year, 2);
    endian_swab2(buf + 12, 1);

    buf[14] = sec1->month;
    buf[15] = sec1->day;
    buf[16] = sec1->hour;
    buf[17] = sec1->min;
    buf[18] = sec1->sec;
    buf[19] = sec1->status;
    buf[20] = sec1->kind;

    fwrite(buf, 1, sec1->length, fp);

#ifdef DEBUG
    grib2_print_sec1(sec1);
#endif

    return 0;
}

static int grib2_write_sec2(const grib2_t *grib2){
    N_UI1 buf[BUF_SIZE];
    sec2_t *sec2 = grib2->data.list->sec2;
    FILE *fp = grib2->fp;

    if (sec2 == NULL) return 0;
    memcpy(buf, &sec2->length, 4);
    endian_swab4(buf, 1);

    buf[4] = sec2->section;

    fwrite(buf, 1, 4, fp);
#ifdef DEBUG
    grib2_print_sec2(sec2);
#endif

    return 0;
}


static int grib2_write_sec3(const grib2_t *grib2){
    N_UI1 buf[BUF_SIZE];
    sec3_t *sec3 = grib2->data.list->sec3;
    FILE *fp = grib2->fp;

    buf[4] = sec3->section;
    buf[5] = sec3->source;

    memcpy(buf + 6, &sec3->n_grid, 4);
    endian_swab4(buf + 6, 1);

    buf[10] = sec3->oct_optlist;
    buf[11] = sec3->list_kind;
    memcpy(buf + 12, &sec3->grid_template, 2);
    endian_swab2(buf + 12, 1);

    if (sec3->grid_template == 30){
      grib2_write_sec3_ln(buf, sec3->proj);
    } else if (sec3->grid_template == 0){
      grib2_write_sec3_ll(buf, sec3->proj);
    }

    memcpy(buf, &sec3->length, 4);
    endian_swab4(buf, 1);

    fwrite(buf, 1, sec3->length, fp);

#ifdef DEBUG
    grib2_print_sec3(sec3);
#endif


    return 0;
}


static int grib2_write_sec3_ln(N_UI1* buf, const sec3_proj_30_t *sec3_ln){
    buf[14] = sec3_ln->earth_shape;

    buf[15] = sec3_ln->r_earth_unit;
    memcpy(buf + 16, &sec3_ln->r_earth, 4);
    endian_swab4(buf + 16, 1);

    buf[20] = sec3_ln->a_earth_unit;
    memcpy(buf + 21,  &sec3_ln->a_earth, 4);
    endian_swab4(buf + 21, 1);

    buf[25] = sec3_ln->b_earth_unit;
    memcpy(buf + 26,  &sec3_ln->b_earth, 4);
    endian_swab4(buf + 26, 1);

    memcpy(buf + 30, &sec3_ln->nx, 4);
    endian_swab4(buf + 30, 1);

    memcpy(buf + 34, &sec3_ln->ny, 4);
    endian_swab4(buf + 34, 1);

    memcpy(buf + 38, &sec3_ln->la1, 4);
    endian_swab4(buf + 38, 1);

    memcpy(buf + 42, &sec3_ln->lo1, 4);
    endian_swab4(buf + 42, 1);

    buf[46] = sec3_ln->res;    

    memcpy(buf + 47, &sec3_ln->lad, 4);
    endian_swab4(buf + 47, 1);

    memcpy(buf + 51, &sec3_ln->lov, 4);
    endian_swab4(buf + 51, 1);

    memcpy(buf + 55, &sec3_ln->dx, 4);
    endian_swab4(buf + 55, 1);

    memcpy(buf + 59, &sec3_ln->dy, 4);
    endian_swab4(buf + 59, 1);

    buf[63] = sec3_ln->flag_center;
    buf[64] = sec3_ln->scan_mode;

    memcpy(buf + 65, &sec3_ln->latin1, 4);
    endian_swab4(buf + 65, 1);

    memcpy(buf + 69, &sec3_ln->latin2, 4);
    endian_swab4(buf + 69, 1);
    
    memcpy(buf + 73, &sec3_ln->proj_sp_lat, 4);
    endian_swab4(buf + 73, 1);

    memcpy(buf + 77, &sec3_ln->proj_sp_lon, 4);
    endian_swab4(buf + 77, 1);

    return 0;
}

static int grib2_write_sec3_ll(N_UI1* buf, const sec3_proj_0_t *sec3_ll){
    buf[14] = sec3_ll->earth_shape;

    buf[15] = sec3_ll->r_earth_unit;
    memcpy(buf + 16, &sec3_ll->r_earth, 4);
    endian_swab4(buf + 16, 1);

    buf[20] = sec3_ll->a_earth_unit;
    memcpy(buf + 21, &sec3_ll->a_earth, 4);
    endian_swab4(buf + 21, 1);

    buf[25] = sec3_ll->b_earth_unit;
    memcpy(buf + 26, &sec3_ll->b_earth, 4);
    endian_swab4(buf + 26, 1);

    memcpy(buf + 30, &sec3_ll->nx, 4);
    endian_swab4(buf + 30, 1);

    memcpy(buf + 34, &sec3_ll->ny, 4);
    endian_swab4(buf + 34, 1);

    memcpy(buf + 38, &sec3_ll->basic_angle, 4);
    endian_swab4(buf + 38, 1);

    memcpy(buf + 42, &sec3_ll->subdivision, 4);
    endian_swab4(buf + 42, 1);

    memcpy(buf + 46, &sec3_ll->la1, 4);
    endian_swab4(buf + 46, 1);

    memcpy(buf + 50, &sec3_ll->lo1, 4);
    endian_swab4(buf + 50, 1);

    buf[54] = sec3_ll->res;    

    memcpy(buf + 55, &sec3_ll->la2, 4);
    endian_swab4(buf + 55, 1);

    memcpy(buf + 59, &sec3_ll->lo2, 4);
    endian_swab4(buf + 59, 1);

    memcpy(buf + 63, &sec3_ll->dx, 4);
    endian_swab4(buf + 63, 1);

    memcpy(buf + 67, &sec3_ll->dy, 4);
    endian_swab4(buf + 67, 1);

    buf[71] = sec3_ll->scan_mode;

    return 0;

}

static int grib2_write_sec4(const grib2_t *grib2){
    N_UI1 buf[BUF_SIZE];
    FILE *fp = grib2->fp;
    sec4_t *sec4 = grib2->data.list->sec4;

    buf[4] = sec4->section;

    memcpy(buf + 5, &sec4->n_grid_after_temp, 2);
    endian_swab2(buf + 5, 1);

    memcpy(buf + 7, &sec4->prod_template, 2);
    endian_swab2(buf + 7, 1);
    
    if (sec4->prod_template == 0){
      grib2_write_sec4_prod_0(buf, sec4->prod_def);
    } else if (sec4->prod_template == 8){
      grib2_write_sec4_prod_8(buf, sec4->prod_def);
    }

    memcpy(buf, &sec4->length, 4);
    endian_swab4(buf, 1);

    fwrite(buf, 1, sec4->length, fp);

#ifdef DEBUG
    grib2_print_sec4(sec4);
#endif

    return 0;
}



static int grib2_write_sec4_prod_0(N_UI1 buf[], const sec4_prod_0_t *sec4_prod_0){
    N_UI4 term;
    
    buf[9] = sec4_prod_0->param_category;
    buf[10] = sec4_prod_0->param_number;
    buf[11] = sec4_prod_0->kind_process;
    buf[12] = sec4_prod_0->bg_process;
    buf[13] = sec4_prod_0->process;

    memcpy(buf + 14, &sec4_prod_0->cutoff_time_hour, 2);
    endian_swab2(buf + 14, 1);

    buf[16] = sec4_prod_0->cutoff_time_min;
    buf[17] = sec4_prod_0->term_unit;

    if (sec4_prod_0->term < 0){
      term = abs(sec4_prod_0->term) | 0x80000000;
    } else {
      term = sec4_prod_0->term;
    }

    memcpy(buf + 18, &term, 4);
    endian_swab4(buf + 18, 1);

    buf[22] = sec4_prod_0->fixed_layer1_kind;
    buf[23] = sec4_prod_0->fixed_layer1_unit;

    memcpy(buf + 24, &sec4_prod_0->fixed_layer1, 4);
    endian_swab4(buf + 24, 1);

    buf[28] = sec4_prod_0->fixed_layer2_kind;
    buf[29] = sec4_prod_0->fixed_layer2_unit;

    memcpy(buf + 30, &sec4_prod_0->fixed_layer2, 4);
    endian_swab4(buf + 30, 1);

    return 0;
}

static int grib2_write_sec4_prod_8(N_UI1 buf[], const sec4_prod_8_t *sec4_prod_8){
    int i;
    N_UI4 term;
    
    buf[9] = sec4_prod_8->param_category;
    buf[10] = sec4_prod_8->param_number;
    buf[11] = sec4_prod_8->kind_process;
    buf[12] = sec4_prod_8->bg_process;
    buf[13] = sec4_prod_8->process;

    memcpy(buf + 14, &sec4_prod_8->cutoff_time_hour, 2);
    endian_swab2(buf + 14, 1);

    buf[16] = sec4_prod_8->cutoff_time_min;
    buf[17] = sec4_prod_8->term_unit;

    if (sec4_prod_8->term < 0){
      term = abs(sec4_prod_8->term) | 0x80000000;
    } else {
      term = sec4_prod_8->term;
    }

    memcpy(buf + 18, &term, 4);
    endian_swab4(buf + 18, 1);

    buf[22] = sec4_prod_8->fixed_layer1_kind;
    buf[23] = sec4_prod_8->fixed_layer1_unit;

    memcpy(buf + 24, &sec4_prod_8->fixed_layer1, 4);
    endian_swab4(buf + 24, 1);

    buf[28] = sec4_prod_8->fixed_layer2_kind;
    buf[29] = sec4_prod_8->fixed_layer2_unit;

    memcpy(buf + 30, &sec4_prod_8->fixed_layer2, 4);
    endian_swab4(buf + 30, 1);

    memcpy(buf + 34, &sec4_prod_8->eyear, 2);
    endian_swab2(buf + 34, 1);

    buf[36] = sec4_prod_8->emonth;
    buf[37] = sec4_prod_8->eday;
    buf[38] = sec4_prod_8->ehour;
    buf[39] = sec4_prod_8->emin;
    buf[40] = sec4_prod_8->esec;
    buf[41] = sec4_prod_8->n;

    memcpy(buf + 42, &sec4_prod_8->n_miss, 4);
    endian_swab4(buf + 42, 1);

    for (i = 0; i < sec4_prod_8->n; i++){
      buf[46 + 12 * i] = sec4_prod_8->stat[i]->stat_process;
      buf[47 + 12 * i] = sec4_prod_8->stat[i]->delt_kind;
      buf[48 + 12 * i] = sec4_prod_8->stat[i]->term_unit;

      if (sec4_prod_8->stat[i]->term < 0){
	term = abs(sec4_prod_8->stat[i]->term) | 0x80000000;	
      } else {
	term = sec4_prod_8->stat[i]->term;
      }

      memcpy(buf + 49 + 12 * i, &term, 4);
      endian_swab4(buf + 49 + 12 * i, 1);

      buf[53 + 12 * i] = sec4_prod_8->stat[i]->delt_unit;

      memcpy(buf + 54 + 12 * i, &sec4_prod_8->stat[i]->delt, 4);
      endian_swab4(buf + 54 + 12 * i, 1);
    }

    return 0;
}




static int grib2_write_sec5(const grib2_t *grib2){
    N_UI1 buf[BUF_SIZE];
    FILE *fp = grib2->fp;
    sec5_t *sec5 = grib2->data.list->sec5;

    buf[4] = sec5->section;
    
    memcpy(buf + 5,  &sec5->n_point, 4);
    endian_swab4(buf + 5, 1);
    
    memcpy(buf + 9, &sec5->rep_template, 2);
    endian_swab2(buf + 9, 1);

    if (sec5->rep_template == 0){
      grib2_write_sec5_rep_0(buf, sec5->rep);
    }
    memcpy(buf, &sec5->length, 4);
    endian_swab4(buf, 1);

    fwrite(buf, 1, sec5->length, fp);

#ifdef DEBUG
    grib2_print_sec5(sec5);
#endif


    return 0;
}

static int grib2_write_sec5_rep_0(N_UI1 *buf, const sec5_rep_0_t *sec5_rep_0){

  N_UI2 unit;

  memcpy(buf + 11, &sec5_rep_0->reference, 4);
  endian_swab4(buf + 11, 1);

  if (sec5_rep_0->binary_unit < 0){
    unit = abs(sec5_rep_0->binary_unit) | 0x8000;
  } else {
    unit = sec5_rep_0->binary_unit;
  }
  memcpy(buf + 15, &unit, 2);
  endian_swab2(buf + 15, 1);

  if (sec5_rep_0->decimal_unit < 0){
    unit = abs(sec5_rep_0->decimal_unit) | 0x8000;
  } else {
    unit = sec5_rep_0->decimal_unit;
  }
  memcpy(buf + 17, &unit, 2);
  endian_swab2(buf + 17, 1);

  buf[19] = sec5_rep_0->n_bit;
  buf[20] = sec5_rep_0->kind_value;

  return 0;
}


static int grib2_write_sec6(const grib2_t *grib2){
  N_UI1 buf[BUF_SIZE];
  FILE *fp = grib2->fp;
  sec6_t *sec6 = grib2->data.list->sec6;
  N_UI1 *bitmap = sec6->bitmap;
  N_UI4 size = sec6->length - 6;

  memcpy(buf, &sec6->length, 4);
  endian_swab4(buf, 1);

  buf[4] = sec6->section;
  buf[5] = sec6->bitmap_flag;

  fwrite(buf, 1, 6, fp);

  if (sec6->bitmap_flag == 0){
    fwrite(bitmap, 1, size, fp);
  } 


#ifdef DEBUG
    grib2_print_sec6(sec6);
#endif

  return 0;
}


static int grib2_write_sec7(const grib2_t *grib2){
  N_UI1 buf[BUF_SIZE];
  FILE *fp = grib2->fp;
  sec7_t *sec7 = grib2->data.list->sec7;
  N_UI1 *data_bit = sec7->value;
  N_UI4 size = sec7->length - 5;

  memcpy(buf, &sec7->length, 4);
  endian_swab4(buf, 1);
  buf[4] = sec7->section;

  fwrite(buf, 1, 5, fp);
  fwrite(data_bit, 1, size, fp);

#ifdef DEBUG
    grib2_print_sec7(sec7);
#endif

  return 0;

}


int grib2_write_end(const grib2_t *grib2){
  N_UI8 total_len;
  FILE *wfp = grib2->fp;

  fwrite("7777", 1, 4, wfp);
  total_len = ftell(wfp);

  fseek(wfp, 8, SEEK_SET);
  endian_swab8(&total_len, 1);
  fwrite(&total_len, 1, 8, wfp);
  /* back to the original position */
  fseek(wfp, total_len, SEEK_SET);

  printf("Output to %s\n", grib2->fname);
  return 0;
}


static int grib2_write_sec3_init(grib2_t *grib2){
  sec3_t *sec3 = grib2->data.list->sec3;

  sec3->proj = NULL;
  return 0;
}

static int grib2_write_sec3_free(grib2_t *grib2){
  sec3_t *sec3 = grib2->data.list->sec3;

  free(sec3->proj);
  sec3->proj = NULL;
  return 0;
}

static int grib2_write_sec4_init(grib2_t *grib2){
  sec4_t *sec4 = grib2->data.list->sec4;

  sec4->prod_def = NULL;
  return 0;
}

static int grib2_write_sec4_free(grib2_t *grib2){
  sec4_t *sec4 = grib2->data.list->sec4;
  sec4_prod_8_t *prod_8;
  int i;

  if (sec4->prod_def != NULL){
    if (sec4->prod_template == 8 || sec4->prod_template == 50008){
      prod_8 = (sec4_prod_8_t*)(sec4->prod_def);
      for (i = 0; i < prod_8->n; i++){
	free(prod_8->stat[i]);
      }
      free(prod_8->stat);
    }
  }
  free(sec4->prod_def);
  sec4->prod_def = NULL;
  return 0;
}

static int grib2_write_sec5_init(grib2_t *grib2){
  sec5_t *sec5 = grib2->data.list->sec5;

  sec5->rep = NULL;
  return 0;
}

static int grib2_write_sec5_free(grib2_t *grib2){
  sec5_t *sec5 = grib2->data.list->sec5;

  sec5->rep = NULL;
  return 0;
}

static int grib2_write_sec6_init(grib2_t *grib2){
  sec6_t *sec6 = grib2->data.list->sec6;

  sec6->bitmap = NULL;
  return 0;
}

static int grib2_write_sec6_free(grib2_t *grib2){
  sec6_t *sec6 = grib2->data.list->sec6;

  free(sec6->bitmap);
  sec6->bitmap = NULL;
  return 0;
}

static int grib2_write_sec7_init(grib2_t *grib2){
  sec7_t *sec7 = grib2->data.list->sec7;

  sec7->value = NULL;
  return 0;
}

static int grib2_write_sec7_free(grib2_t *grib2){
  sec7_t *sec7 = grib2->data.list->sec7;

  free(sec7->value);
  sec7->value = NULL;
  return 0;
}

static int grib2_get_earth_rad(const N_UI1 earth_shape, 
			       N_UI1 *r_earth_unit, N_UI4 *r_earth, 
			       N_UI1 *a_earth_unit, N_UI4 *a_earth, 
			       N_UI1 *b_earth_unit, N_UI4 *b_earth){

  switch(earth_shape){
  case (EARTH_SHAPE_SPHERE_USERRAD):
    *r_earth_unit = 0;
    *a_earth_unit = 0xff;
    *b_earth_unit = 0xff;

    *r_earth = floor(nwp_get_earth_rad_m() * pow(10.0, (*r_earth)) + 0.5);    
    *a_earth = 0xffffffff;
    *b_earth = 0xffffffff;
    
    break;
  case(EARTH_SHAPE_GRS80):
    *r_earth_unit = 0xff;
    *a_earth_unit = 1;
    *b_earth_unit = 1;

    *r_earth = 0xffffffff;
    *a_earth = floor(nwp_get_earth_rad_e() * pow(10.0, (*a_earth)) + 0.5);
    *b_earth = floor(nwp_get_earth_rad_p() * pow(10.0, (*b_earth)) + 0.5);
    break;
  }
  return 0;
}

static int grib2_write_sec3_set(const N_UI1 earth_shape, 
				const grid_info_t *gi, sec3_t *sec3){
  sec3_proj_0_t *proj_0;
  sec3_proj_30_t *proj_30;
  double fi;
  double fj;
  double flat1;
  double flon1;

  sec3->section = 3;
  sec3->source = 0;
  sec3->n_grid = gi->nxy;
  sec3->oct_optlist = 0;
  sec3->list_kind = 0;

  if (memcmp(gi->proj, "LL", 2)  == 0){
    sec3->length = 72;

    proj_0 = (sec3_proj_0_t*)malloc(sizeof(sec3_proj_0_t));
    sec3->grid_template = 0; /* lat-lon */
    sec3->proj = proj_0;

    proj_0->earth_shape = earth_shape;
    
    grib2_get_earth_rad(earth_shape, 
			&(proj_0->r_earth_unit), 
			&(proj_0->r_earth), 
			&(proj_0->a_earth_unit), 
			&(proj_0->a_earth), 
			&(proj_0->a_earth_unit), 
			&(proj_0->b_earth));

    proj_0->nx = gi->nx;
    proj_0->ny = gi->ny;
    proj_0->basic_angle = 0;
    proj_0->subdivision = 0xffffffff;
    proj_0->la1 = floor((gi->xlat + (gi->xj - 1.0) * gi->dy) * 1.0e6 + 0.5);
    proj_0->lo1 = floor((gi->xlon - (gi->xi - 1.0) * gi->dx) * 1.0e6 + 0.5);
    proj_0->res = 48;
    proj_0->la2 = floor((gi->xlat - ((double)gi->ny - gi->xj) * gi->dy) * 1.0e6 + 0.5);
    proj_0->lo2 = floor((gi->xlon + ((double)gi->nx - gi->xi) * gi->dx) * 1.0e6 + 0.5);
    proj_0->dx = floor(gi->dx * 1.0e6 + 0.5);
    proj_0->dy = floor(gi->dy * 1.0e6 + 0.5);
    proj_0->scan_mode = 0;
  } else if (memcmp(gi->proj, "LM", 2) == 0){
    sec3->length = 81;
    proj_30 = (sec3_proj_30_t*)malloc(sizeof(sec3_proj_30_t));
    sec3->grid_template = 30; /* Lambert */
    sec3->proj = proj_30;

    proj_30->earth_shape = earth_shape;

    grib2_get_earth_rad(earth_shape, 
			&(proj_30->r_earth_unit), 
			&(proj_30->r_earth), 
			&(proj_30->a_earth_unit), 
			&(proj_30->a_earth), 
			&(proj_30->a_earth_unit), 
			&(proj_30->b_earth));

    proj_30->nx = gi->nx;
    proj_30->ny = gi->ny;

    fi = 1.0;
    fj = 1.0;

    NWP_lambert2sphere_D(&fi, &fj, 1, 
                         gi->slat1, gi->slat2, gi->slon, 
                         gi->xlat, gi->xlon, gi->xi, gi->xj, 
                         gi->dx, &flat1, &flon1);

    proj_30->la1 = floor(flat1 * 1.0e6 + 0.5);
    proj_30->lo1 = floor(flon1 * 1.0e6 + 0.5);
    proj_30->res = 8; 
    proj_30->lad = floor(gi->xlat * 1.0e6 + 0.5);
    proj_30->lov = floor(gi->xlon * 1.0e6 + 0.5);
    proj_30->dx = floor(gi->dx * 1.0e3 + 0.5);
    proj_30->dy = floor(gi->dy * 1.0e3 + 0.5);
    proj_30->flag_center = 0;
    proj_30->scan_mode = 0;
    proj_30->latin1 = floor(gi->slat2 * 1.0e6 + 0.5);
    proj_30->latin2 = floor(gi->slat1 * 1.0e6 + 0.5);
    proj_30->proj_sp_lat = 0xffffffff;
    proj_30->proj_sp_lon = 0xffffffff;

  } else {
    fprintf(stderr, "Not supported projection\n");
    return ERR_UNSUPPORTED_PROJ_TEMPLATE;
  }
  return 0;
}


static int grib2_write_sec4_set(const N_UI2 prod_template, 
			 const N_UI1 parm_cat, 
			 const N_UI1 parm_num, 
			 const N_UI2 cutoff_time_hour, 
			 const N_UI2 cutoff_time_min, 
			 const N_SI4 term, 
			 const N_UI1 layer_kind, 
			 const N_UI1 layer_unit, const N_UI4 layer, 
			 const ymdhm_t *eymdhm, 
			 const N_UI1 n_stat, const N_UI4 *stat_term, 
			 sec4_t *sec4){

  sec4_prod_0_t *prod_0;
  sec4_prod_8_t *prod_8;
  int i;

  sec4->section = 4;
  sec4->n_grid_after_temp = 0;
  sec4->prod_template = prod_template;

  switch (prod_template){
  case (0):
    sec4->length = 34;

    prod_0 = (sec4_prod_0_t*)malloc(sizeof(sec4_prod_0_t));
    sec4->prod_def = prod_0;

    prod_0->param_category = parm_cat;
    prod_0->param_number = parm_num;
    prod_0->kind_process = 0;
    prod_0->bg_process = 31;
    prod_0->process = 0xff;
    prod_0->cutoff_time_hour = cutoff_time_hour;
    prod_0->cutoff_time_min = cutoff_time_min;
    prod_0->term_unit = 0;
    prod_0->term = term;
    prod_0->fixed_layer1_kind = layer_kind;
    if (layer_kind == LAYER_SURFACE || layer_kind == LAYER_MSL){
      prod_0->fixed_layer1_unit = 0xff;
      prod_0->fixed_layer1 = 0xffffffff;
    } else {
      prod_0->fixed_layer1_unit = layer_unit;
      prod_0->fixed_layer1 = layer;
    }
    prod_0->fixed_layer2_kind = 0xff;
    prod_0->fixed_layer2_unit = 0xff;
    prod_0->fixed_layer2 = 0xffffffff;
    break;
  case (8):
  case (50008):
    sec4->length = 46 + 12 * n_stat;
    prod_8 = (sec4_prod_8_t*)malloc(sizeof(sec4_prod_8_t));
    sec4->prod_def = prod_8;
    prod_8->param_category = parm_cat;
    prod_8->param_number = parm_num;
    prod_8->kind_process = 0;
    prod_8->bg_process = 31;
    prod_8->process = 0;
    prod_8->cutoff_time_hour = cutoff_time_hour;
    prod_8->cutoff_time_min = cutoff_time_min;
    prod_8->term_unit = 0;
    prod_8->term = term;
    prod_8->fixed_layer1_kind = layer_kind;
    if (layer_kind == LAYER_SURFACE || layer_kind == LAYER_MSL){
      prod_8->fixed_layer1_unit = 0xff;
      prod_8->fixed_layer1 = 0xffffffff;
    } else {
      prod_8->fixed_layer1_unit = layer_unit;
      prod_8->fixed_layer1 = layer;
    }
    prod_8->fixed_layer2_kind = 0xff;
    prod_8->fixed_layer2_unit = 0xff;
    prod_8->fixed_layer2 = 0xffffffff;
    prod_8->eyear = eymdhm->year;
    prod_8->emonth = eymdhm->month;
    prod_8->eday = eymdhm->day;
    prod_8->ehour = eymdhm->hour;
    prod_8->emin = eymdhm->min;
    prod_8->esec = 0;
    prod_8->n = n_stat;
    prod_8->n_miss = 0;

    prod_8->stat = (sec4_prod_8_stat_t**)
                     malloc(n_stat * sizeof(sec4_prod_8_stat_t*));
    for (i = 0; i < n_stat; i++){
      prod_8->stat[i] = (sec4_prod_8_stat_t*)
	                 malloc(sizeof(sec4_prod_8_stat_t));
      prod_8->stat[i]->stat_process = 1;
      prod_8->stat[i]->delt_kind = 2;
      prod_8->stat[i]->term_unit = 0;
      prod_8->stat[i]->term = stat_term[i]; /* minute */
      prod_8->stat[i]->delt_unit = 0;
      prod_8->stat[i]->delt = 0; /* assume */
    }

    /* only for 50008 */
    /* however, the additional infomation is not written to 
     converted grib2. */
    /*
    if (prod_template == 50008){
      sec4->length += 24;
      prod_8->radar_use1 = radar_use1;
      prod_8->radar_use2 = radar_use2;
      prod_8->raingauge_use = raingauge_use;
    }
    */
    break;
  default:
    fprintf(stderr, "Not supported product template\n");
    return ERR_UNSUPPORTED_PROD_TEMPLATE;
    break;
  }

  return 0;
}


static int grib2_write_sec5_set(const N_UI4 n_point, const N_UI2 rep_template,
			 const float ref, const N_SI2 binary_unit, 
			 const N_SI2 decimal_unit, N_UI1 n_bit, 
			 sec5_t *sec5){
  sec5_rep_0_t *rep_0;

  /* only rep_template = 0 is supported */
  sec5->section = 5;
  sec5->n_point = n_point;
  sec5->rep_template = rep_template;

  switch(rep_template){
  case(0):
    sec5->length = 21;
    rep_0 = (sec5_rep_0_t*)malloc(sizeof(sec5_rep_0_t));
    sec5->rep = rep_0;

    rep_0->reference = ref;
    rep_0->binary_unit = binary_unit;
    rep_0->decimal_unit = decimal_unit;
    rep_0->n_bit = n_bit;
    break;
  default:
    fprintf(stderr, "Not supported product template\n");
    return ERR_UNSUPPORTED_REP_TEMPLATE;
    break;
  }
  return 0;
}

#define EPS_REL 1.0e-10

static int grib2_write_sec567_gen_bm_value(const N_UI4 n_grid, 
			     const float *data_in, 
			     N_UI4 *n_point, N_UI1 *bitmap, float *data_out){
  N_UI4 j = 0;
  N_UI4 bm_size = (n_grid - 1) / 8 + 1;
  N_UI4 i;

  for (i = 0; i < bm_size; i++){
    bitmap[i] = 0;
  }

  for (i = 0; i < n_grid; i++){
    if (data_in[i] == MISSING){
    } else {
      data_out[j++] = data_in[i];
      bitmap[i / 8] |= (1 << (7 - i % 8));
    }
  }
  *n_point = j;
  return 0;

}

static int grib2_write_sec6_set(const N_UI4 n_grid, const N_UI4 n_point, 
			 N_UI1 *bitmap, sec6_t *sec6){

  sec6->section = 6;
  if (n_grid == n_point){
    sec6->bitmap_flag = 0xff;
    sec6->bitmap = NULL;
    sec6->length = 6;
  } else {
    sec6->bitmap_flag = 0;
    sec6->bitmap = bitmap;
    sec6->length = 6 + (n_grid - 1) / 8 + 1;
  }
  return 0;
}

static int grib2_write_sec7_set(N_UI1 *data_bit, const N_UI4 dsize, 
			 sec7_t *sec7){
  sec7->section = 7;
  sec7->value = data_bit;
  sec7->length = 5 + dsize;

  return 0;
}

int grib2_write_sec0123(const N_UI1 earth_shape, 
			const grid_info_t *gi, grib2_t *grib2_in, 
			grib2_t *grib2_wrt){
  sec3_t *sec3 = grib2_wrt->data.list->sec3;
  /* point to the first record */
  grib2_wrt->data.list->sec0 = grib2_in->data.list->sec0;
  grib2_wrt->data.list->sec1 = grib2_in->data.list->sec1;
  grib2_wrt->data.list->sec2 = grib2_in->data.list->sec2;

  grib2_write_sec3_init(grib2_wrt);

  grib2_write_sec3_set(earth_shape, gi, sec3);

  grib2_write_sec0(grib2_wrt);
  grib2_write_sec1(grib2_wrt);
  grib2_write_sec2(grib2_wrt);
  grib2_write_sec3(grib2_wrt);

  grib2_write_sec3_free(grib2_wrt);

  return 0;
}

#define EPS 1.0e-6
#define ENCODE_NBIT 12

int grib2_write_sec4567(const char *elem_in, 
			const N_UI2 cutoff_time_hour, 
			const N_UI1 cutoff_time_min,
			const N_SI4 term, 
			const N_UI4 layer_kind, const double *layer_in, 
			const N_UI4 nz, 
			ymdhm_t *eymdhm_in, 
			const float *data_in, 
			grib2_t *grib2_wrt){
  char elem[6];
  N_UI4 len;
  N_UI1 parm_cat;
  N_UI1 parm_num;
  N_UI2 prod_template;
  N_UI1 n_stat;
  N_UI4 stat_term;
  N_UI4 packed_len;
  const N_UI2 rep_template = 0;
  N_SI2 decimal_unit;
  N_SI2 binary_unit;
  float reference;
  N_UI1 n_bit = ENCODE_NBIT;
  N_UI1 *uc;
  N_UI4 uc_size;
  N_UI4 n_point;
  float *data_out;
  N_UI1 *bitmap;
  ymdhm_t *eymdhm;
  N_UI1 layer_unit;
  double layer_in_tmp;
  N_UI4 layer;
  N_UI4 n_grid;
  N_UI4 kk;

  grib2_write_sec4_init(grib2_wrt);
  grib2_write_sec5_init(grib2_wrt);
  grib2_write_sec6_init(grib2_wrt);
  grib2_write_sec7_init(grib2_wrt);


  len = strlen(elem_in);
  memcpy(elem, elem_in, len);
  if (len < 6){
    memset(elem + len, ' ', 6 - len);
  }

  grib2_get_parm(elem, &parm_cat, &parm_num);
  if (memcmp(elem, "RAIN  ", 6) == 0 || memcmp(elem, "RR60  ", 6) == 0){
    prod_template = 8;
    eymdhm = eymdhm_in;
    n_stat = 1;
    stat_term = 60;
  } else {
    prod_template = 0;
    eymdhm = NULL;
    n_stat = 0;
    stat_term = 0; /* dummy, not used */
  }

  fprintf(stderr, "GRIB2 processing: %s\n", elem_in);
  for (kk = 0; kk < nz; kk++){
    if (layer_in != NULL){
      layer_unit = 0;
      layer_in_tmp = layer_in[kk];

      while (fabs((N_UI4)(floor(layer_in_tmp)) - layer_in_tmp) > EPS){
	layer_in_tmp = layer_in_tmp * 10;
	layer_unit++;
      }
      layer = (N_UI4)(layer_in_tmp);
    } else {
      layer_unit = 0xff;
      layer = 0xffffffff;
    }

    grib2_write_sec4_set(prod_template, 
			 parm_cat, 
			 parm_num, 
			 cutoff_time_hour, 
			 cutoff_time_min, 
			 term, 
			 layer_kind, 
			 layer_unit, layer, 
			 eymdhm, 
			 n_stat, &stat_term, 
			 grib2_wrt->data.list->sec4);

    n_grid = grib2_wrt->data.list->sec3->n_grid;
    /* alloc maxium size*/
    bitmap = (N_UI1*)malloc(sizeof(N_UI1) * ((n_grid - 1) / 8 + 1));
    data_out = (float*)malloc(sizeof(float) * n_grid);

    grib2_write_sec567_gen_bm_value(n_grid, 
				    data_in + n_grid * kk, 
				    &n_point, 
				    bitmap, 
				    data_out);
    
    uc_size = (n_bit * n_point - 1) / 8 + 1 + 100;
    uc = (N_UI1*)malloc(uc_size);

    decimal_unit = 0;
    grib2_encode_rep_0(n_point, n_bit, decimal_unit,
		       data_out, 
		       &binary_unit,
		       &reference,
		       uc, uc_size, &packed_len);
    free(data_out);

    grib2_write_sec5_set(n_point, rep_template,
			 reference, binary_unit, 
			 decimal_unit, n_bit, 
			 grib2_wrt->data.list->sec5);


    grib2_write_sec6_set(n_grid, n_point, 
			 bitmap, grib2_wrt->data.list->sec6);

    grib2_write_sec7_set(uc, packed_len, 
			 grib2_wrt->data.list->sec7);


    grib2_write_sec4(grib2_wrt);
    grib2_write_sec5(grib2_wrt);
    grib2_write_sec6(grib2_wrt);
    grib2_write_sec7(grib2_wrt);


    grib2_write_sec4_free(grib2_wrt);
    grib2_write_sec5_free(grib2_wrt);
    grib2_write_sec6_free(grib2_wrt);
    grib2_write_sec7_free(grib2_wrt);
  }

  return 0;
}

