#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "interp.h"

static int interp_alloc(const grid_info_t *gi_out, interp_2d_t *i2d_w);

static int interp_alloc(const grid_info_t *gi_out, interp_2d_t *i2d_w){
  N_UI4 nxy = gi_out->nx * gi_out->ny;

  i2d_w->nx = gi_out->nx;
  i2d_w->ny = gi_out->ny;

  if ((i2d_w->t11 = (double*)malloc(nxy * sizeof(double))) == NULL ||
      (i2d_w->t12 = (double*)malloc(nxy * sizeof(double))) == NULL ||
      (i2d_w->t21 = (double*)malloc(nxy * sizeof(double))) == NULL ||
      (i2d_w->t22 = (double*)malloc(nxy * sizeof(double))) == NULL ||
      (i2d_w->fi = (double*)malloc(nxy * sizeof(double))) == NULL || 
      (i2d_w->fj = (double*)malloc(nxy * sizeof(double))) == NULL){
    fprintf(stderr, "malloc error at lines %d in %s.\n", __LINE__, __FILE__);
    return ERR_MALLOC;
  }
  return 0;

}

static int interp_northdx(const grid_info_t *gi_in, const grid_info_t *gi_out, 
                          const double *fi, const double *fj, double *dnth){
  
  const double pi = PI;
  double poi = pi / 180.0;
  double rpoi = 180.0 / pi;
  double epsd = (gi_in->dx < gi_in->dy ? gi_in->dy : gi_in->dx) * 1.0e-4;
  double pi4 = pi * 0.25;

  double slata;
  double slatb;
  double slata1;
  double slatb1;

  double slat1;
  double slat2;
  double slon;
  double xlat;
  double xlon;
  double xlat1;
  double xlon1;
  double ck;
  double rck;

  double a1 = nwp_get_earth_rad_m();
  double acn;
  double rl0;
  double xslon;
  double x0;
  double y0;
  double xp;
  double yp;
  
  int ixy;
  int i;
  int j;


  if (memcmp(gi_in->proj, "LM", 2) == 0){
    slata = gi_in->slat1;
    slatb = gi_in->slat2;
    xlat = gi_in->xlat;
    xlon = gi_in->xlon;
    slon = gi_in->slon;

    slata1 = slata * poi;
    slatb1 = slatb * poi;
    slat1 = pi4 - slata * poi * 0.5;
    slat2 = pi4 - slatb * poi * 0.5;
    xlat1 = xlat * poi;
    xlon1 = xlon * poi;
    ck = log(cos(slata1) / cos(slatb1)) / log(tan(slat1) / tan(slat2));
    rck = 1.0 / ck;
    acn = a1 * cos(slata1) * rck;
    rl0 = acn * pow(tan(pi4 - xlat1 * 0.5) / tan(slat1), ck);

    xslon = xlon - slon;
    xslon = fmod(xslon + 900.0, 360.0) - 180.0;
    xslon = xslon * poi * ck;
    x0 = rl0 * sin(xslon);
    y0 = rl0 * cos(xslon);

    if (gi_in == gi_out && fi == NULL && fj == NULL){
      for (ixy = 0; ixy < gi_in->nxy; ixy++){
        i = ixy % gi_in->nx + 1;
        j = ixy / gi_in->nx + 1;

        xp = x0 + (i - gi_in->xi) * gi_in->dx;
        yp = y0 + (j - gi_in->xj) * gi_in->dy;

        if (fabs(xp) < epsd && fabs(yp) < epsd){
          dnth[ixy] = 0.0;
        } else {
          dnth[ixy] = rpoi * atan2(xp, yp);
        }
      }
    } else {
      for (ixy = 0; ixy < gi_out->nxy; ixy++){
        xp = x0 + (fi[ixy] - gi_in->xi) * gi_in->dx;
        yp = y0 + (fj[ixy] - gi_in->xj) * gi_in->dy;

        if (fabs(xp) < epsd && fabs(yp) < epsd){
          dnth[ixy] = 0.0;
        } else {
          dnth[ixy] = rpoi * atan2(xp, yp);
        }
      }
    }
  } else if (memcmp(gi_in->proj, "LL", 2) == 0){
    for (ixy = 0; ixy < gi_out->nxy; ixy++){
      dnth[ixy] = 0.0;
    }    
  } else {
    fprintf(stderr, "wndrot does not support the projection.\n");
    return ERR_UNKNOWN_GRIDINFO_PROJ;
  }

  return 0;

}


int interp_prepare(const grid_info_t *gi_in, const grid_info_t *gi_out, 
		   const N_UI1 allow_out_domain, interp_2d_t *i2d_w){

  int ixy;
  int r;

  double xx;
  double yy;
  int ixo;
  int jyo;
  double xo1;
  double xo2;
  double yo1;
  double yo2;

  double *lat_out;
  double *lon_out;

  double *fi_out;
  double *fj_out;

  if((r = interp_alloc(gi_out, i2d_w))) return r;
  
  i2d_w->allow_out_domain = allow_out_domain;

  if ((lat_out = (double*)malloc(gi_out->nxy * sizeof(double))) == NULL || 
      (lon_out = (double*)malloc(gi_out->nxy * sizeof(double))) == NULL){
    fprintf(stderr, "malloc error at lines %d in %s.\n", __LINE__, __FILE__);
    return ERR_MALLOC;
  }
  /* Get Lat and Lon on the output */
  if (memcmp(gi_out->proj, "LL", 2) == 0){
    for (ixy = 0; ixy < gi_out->nxy; ixy++){
      lon_out[ixy] = gi_out->xlon 
        + (ixy % gi_out->nx + 1 - gi_out->xi) * gi_out->dx;
      lat_out[ixy] = gi_out->xlat 
        - (ixy / gi_out->nx + 1 - gi_out->xj) * gi_out->dy;
    }
  } else if (memcmp(gi_out->proj, "LM", 2) == 0){
    fi_out = (double*)malloc(gi_out->nxy * sizeof(double));
    fj_out = (double*)malloc(gi_out->nxy * sizeof(double));

    for (ixy = 0; ixy < gi_out->nxy; ixy++){
      fi_out[ixy] = ixy % gi_out->nx + 1;
      fj_out[ixy] = ixy / gi_out->nx + 1;
    }
    /* Calculate lat and lon on Lambert */
    NWP_lambert2sphere_D(fi_out, fj_out, gi_out->nxy, 
                         gi_out->slat1, gi_out->slat2, gi_out->slon, 
                         gi_out->xlat, gi_out->xlon, gi_out->xi, gi_out->xj, 
                         gi_out->dx, lat_out, lon_out);

    free(fi_out);
    free(fj_out);
  }

  /* get fi and fi of the input on the output*/
  if (memcmp(gi_in->proj, "LM", 2) == 0){
    NWP_sphere2lambert_D(lat_out, lon_out, 
                         gi_out->nxy, gi_in->slat1, gi_in->slat2, 
                         gi_in->slon, gi_in->xlat, gi_in->xlon, 
                         gi_in->xi, gi_in->xj, gi_in->dx,
                         i2d_w->fi, i2d_w->fj);
  } else if (memcmp(gi_in->proj, "LL", 2) == 0){
    for (ixy = 0; ixy < gi_out->nxy; ixy++){
      i2d_w->fi[ixy] = gi_in->xi 
        + (lon_out[ixy] - gi_in->xlon) / gi_in->dx;
      i2d_w->fj[ixy] = gi_in->xj 
        - (lat_out[ixy] - gi_in->xlat) / gi_in->dy;
    }
  }
  free(lon_out);
  free(lat_out);

  for (ixy = 0; ixy < gi_out->nxy; ixy++){
    xx = i2d_w->fi[ixy];
    yy = i2d_w->fj[ixy];

    ixo = floor(xx);
    jyo = floor(yy);

    if (ixo < 1 || ixo >= gi_in->nx || jyo < 1 || jyo >= gi_in->ny){
      if (i2d_w->allow_out_domain == 0){
	fprintf(stderr, "Error! Out of domain.\n");
	return ERR_INTERP_OUT_OF_DOMAIN;
      } else {
	i2d_w->t11[ixy] = MISSING;
	i2d_w->t21[ixy] = MISSING;
	i2d_w->t12[ixy] = MISSING;
	i2d_w->t22[ixy] = MISSING;
      }
    } else {
      xo1 = (double)ixo;
      xo2 = (double)(ixo + 1);
      yo1 = (double)jyo;
      yo2 = (double)(jyo + 1);

      i2d_w->t11[ixy] = (xo2 - xx) * (yo2 - yy);
      i2d_w->t21[ixy] = (xx - xo1) * (yo2 - yy);
      i2d_w->t12[ixy] = (xo2 - xx) * (yy - yo1);
      i2d_w->t22[ixy] = (xx - xo1) * (yy - yo1);
    }
  }

  return 0;
}

void interp_run(const grid_info_t *gi_in, const grid_info_t *gi_out,
		const unsigned int nz, 
		const interp_2d_t *i2d_w, 
		const float data_ln[], float data_ll[]){
  
  int ixy;
  int im;
  int ip;
  int jm;
  int jp;
  int nx_in = gi_in->nx;
  int kz;
  int kz_ind_in;

  for(kz = 0; kz < nz; kz++){
    for (ixy = 0; ixy < gi_out->nxy; ixy++){
      if (i2d_w->t11[ixy] > 1.0e10){
	data_ll[kz * gi_out->nxy + ixy] = MISSING;
      } else {
      /* Note: indexes are different between models 
         (i.e. Fortran, starting from 1) and C (starting from 0) */
	im = floor(i2d_w->fi[ixy]) - 1;
	ip = im + 1;

	jm = floor(i2d_w->fj[ixy]) - 1;
	jp = jm + 1;

	kz_ind_in = kz * gi_in->nxy;

	data_ll[kz * gi_out->nxy + ixy] = 
	  i2d_w->t11[ixy] * data_ln[kz_ind_in  + jm * nx_in + im]
	  + i2d_w->t12[ixy] * data_ln[kz_ind_in + jp * nx_in + im]
	  + i2d_w->t21[ixy] * data_ln[kz_ind_in + jm * nx_in + ip]
	  + i2d_w->t22[ixy] * data_ln[kz_ind_in + jp * nx_in + ip];
      }
    }
  }
  return;

}
void interp_nearest_run(const grid_info_t *gi_in, const grid_info_t *gi_out,
			const unsigned int nz, 
			const interp_2d_t *i2d_w, 
			const float data_ln[], float data_ll[]){
  
  int ixy;
  int im;
  int ip;
  int jm;
  int jp;
  int nx_in = gi_in->nx;
  int kz;
  int kz_ind_in;
  int i_max;
  double max;
  double t11;
  double t12;
  double t21;
  double t22;

  for(kz = 0; kz < nz; kz++){
    for (ixy = 0; ixy < gi_out->nxy; ixy++){
      if (i2d_w->t11[ixy] > 1.0e10){
	data_ll[kz * gi_out->nxy + ixy] = MISSING;
      } else {
	/* Note: indexes are different between models 
	   (i.e. Fortran, starting from 1) and C (starting from 0) */
	im = floor(i2d_w->fi[ixy]) - 1;
	ip = im + 1;

	jm = floor(i2d_w->fj[ixy]) - 1;
	jp = jm + 1;

	max = -1.0e30;
	if (i2d_w->t11[ixy] > max){
	  max = i2d_w->t11[ixy];
	  i_max = 11;
	}
	if (i2d_w->t12[ixy] > max){
	  max = i2d_w->t12[ixy];
	  i_max = 12;
	}
	if (i2d_w->t21[ixy] > max){
	  max = i2d_w->t21[ixy];
	  i_max = 21;
	}
	if (i2d_w->t22[ixy] > max){
	  max = i2d_w->t22[ixy];
	  i_max = 22;
	}

	t11 = 0.0;
	t12 = 0.0;
	t21 = 0.0;
	t22 = 0.0;
	if (i_max == 11){
	  t11 = 1.0;
	} else if (i_max == 12){
	  t12 = 1.0;
	} else if (i_max == 21){
	  t21 = 1.0;
	} else if (i_max == 22){
	  t22 = 1.0;
	}

	kz_ind_in = kz * gi_in->nxy;
	data_ll[kz * gi_out->nxy + ixy] = 
	  t11 * data_ln[kz_ind_in  + jm * nx_in + im]
	  + t12 * data_ln[kz_ind_in + jp * nx_in + im]
	  + t21 * data_ln[kz_ind_in + jm * nx_in + ip]
	  + t22 * data_ln[kz_ind_in + jp * nx_in + ip];
      }
    }
  }
  return;

}

int interp_wndrot(const grid_info_t *gi_in, const grid_info_t *gi_out, 
		  const unsigned int nz,
                  const double *fi, const double *fj, 
                  float *u, float *v){

  double *dnth_in;
  double *dnth_out;
  double uu;
  double vv;
  double rad;
  double deg2rad = PI / 180.0;
  double angs;
  double angc;
  double *angle;
  int ixy;
  int kz;
  int idx;
  int r;

  dnth_in = (double*)malloc(gi_out->nxy * sizeof(double));
  dnth_out = (double*)malloc(gi_out->nxy * sizeof(double));

  if ((r = interp_northdx(gi_out, gi_out, 
			  NULL, NULL, dnth_out))) return r;
  if ((r = interp_northdx(gi_in, gi_out, 
			  fi, fj, dnth_in))) return r;
  
  angle = (double*)malloc(gi_out->nxy * sizeof(double));

  for (ixy = 0; ixy < gi_out->nxy; ixy++){
    angle[ixy] = fmod((dnth_out[ixy] - dnth_in[ixy]) + 900.0, 360.0) - 180.0;
  }
  free(dnth_in);
  free(dnth_out);

  for (kz = 0; kz < nz; kz++){
    for (ixy = 0; ixy < gi_out->nxy; ixy++){
      rad = angle[ixy] * deg2rad;
      angs = sin(rad);
      angc = cos(rad);
      
      idx = kz * gi_out->nxy + ixy;

      uu = angc * u[idx] - angs * v[idx];
      vv = angs * u[idx] + angc * v[idx];
      u[idx] = uu;
      v[idx] = vv;
    }
  }
  free(angle);

  return 0;
}

int interp_mxme(const grid_info_t *gi_in, const grid_info_t *gi_out, 
		const unsigned int ope, 
		const float *data_in, float *data_out){

  interp_2d_t interp_rev;
  unsigned int *n_data_out;
  int i2;
  int j2;
  int idx;
  int ixy;

  /* get fi and fj of gi_out from gi_in */
  /* intentionally reversed gi_in and gi_out */
  interp_prepare(gi_out, gi_in, 1, &interp_rev);

  if ((n_data_out = (unsigned int*)malloc(sizeof(unsigned int) * gi_out->nxy)) == NULL){
    fprintf(stderr, "malloc error at lines %d in %s.\n", __LINE__, __FILE__);
    return ERR_MALLOC;
  }

  for (ixy = 0; ixy < gi_out->nxy; ixy++){
    data_out[ixy] = 0.0;
    n_data_out[ixy] = 0;
  }

  for (ixy = 0; ixy < gi_in->nxy; ixy++){
    i2 = floor(interp_rev.fi[ixy] + 0.5);
    j2 = floor(interp_rev.fj[ixy] + 0.5);

    if (i2 < 1 || i2 > gi_out->nx || j2 < 1 || j2 > gi_out->ny){
      continue;
    }
    idx = (j2 - 1) * gi_out->nx + (i2 - 1);
    if (data_in[ixy] > 1.0e10){
      continue;
    } else if (ope == INTERP_OPE_MEAN){
      data_out[idx] += data_in[ixy];
      n_data_out[idx]++;
    } else if (ope == INTERP_OPE_MAX){
      if (data_in[ixy] > data_out[idx]){
	data_out[idx] = data_in[ixy];
      }
      n_data_out[idx]++;
    }
  }
  for (ixy = 0; ixy < gi_out->nxy; ixy++){
    if (n_data_out[ixy] > 0){
      if (ope == INTERP_OPE_MEAN){
	data_out[ixy] = data_out[ixy] / (double)(n_data_out[ixy]);
      }
    } else {
      data_out[ixy] = MISSING;
    }
  }

  free(n_data_out);
  return 0;
}
