#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "typedef.h"
#include "const.h"
#include "grib2_io.h"
#include "ymdhm.h"
#include "parse_config.h"
#include "parse_opt.h"
#include "z2p.h"
#include "interp.h"
#include "grib2_parm_table.h"
#include "fdat.h"
#include "grads_io.h"
#include "grads_wrt.h"
#include "gpv.h"
#include "nwpl_map_earth.h"

int main(int argc, char *argv[]){

  int conv_h = 0;
  int conv_v = 0;

  const N_UI2 cutoff_time_hour_ma = 0;
  const N_UI1 cutoff_time_min_ma = 50;

  const N_UI2 cutoff_time_hour_ra = 0;
  const N_UI1 cutoff_time_min_ra = 10;

  const N_SI4 term_ma = 0;
  const N_SI4 term_ra = -60; 

  int r;

  double *org_layer = NULL;
  N_UI1 parm_cat;
  N_UI1 parm_num;
  int kk;
  int nz;

  gpv_t gpv;
  grib2_t grib2_in;
  grib2_t grib2_wrt;
  grds_t grds;

  ymdhm_t ymdhm;
  z2p_info_t z2p;
  grid_info_t gi_in;
  grid_info_t gi_out;
  grid_info_t *gi_target;

  opt_t opt;
  config_h_t config_h;
  config_v_t config_v;

  interp_2d_t interp;

  if (argc < 2){
    parse_opt_usage(argv[0]);
    return 1;
  }

  if ((r = parse_opt(argc, argv, &opt))) return r;
  parse_opt_print(&opt);

  if (opt.fname_config_h != NULL){
    if ((r = parse_config_h(opt.fname_config_h, &config_h))){
      fprintf(stderr, "Error occured when reading config_h file: %s.\n", 
              opt.fname_config_h);
      return r;
    } 
    parse_config_print_config_h(&config_h);

    grid_info_set_from_config_h(&config_h, &gi_out);
#ifdef DEBUG
    grid_info_print(&gi_out);
#endif

    conv_h = 1;
  }

  if (opt.fname_config_v != NULL){
    if ((r = parse_config_v(opt.fname_config_v, &config_v))) {
      fprintf(stderr, "Error occured when reading config_v file: %s.\n", 
              opt.fname_config_v);
      return r;
    } 
    parse_config_print_config_v(&config_v);
    conv_v = 1;
  }
  grib2_init(&grib2_in);
  if ((r = grib2_open(opt.fname_grib2, "r", &grib2_in))) return r;
  if ((r = grib2_data_set(&grib2_in))){
    fprintf(stderr, "Failed to read.\n");
    return r;
  }

  if (opt.print_rec == 1){
    grib2_print_rec(&grib2_in);
    return 0;
  }

  if ((r = nwp_set_earth(NWP_EARTH_GRS80))) return r;
  grid_info_set_from_sec3(grib2_in.data.list->sec3, &gi_in);
  grib2_ymdhm_set_from_sec1(grib2_in.data.list->sec1, &ymdhm);

#ifdef DEBUG
  fprintf(stderr, "=== Grid Info of inputted GRIB2 ===\n");
  grid_info_print(&gi_in);
  fprintf(stderr, "rad: %f\n", grib2_inq_earth_rad(grib2_in.data.list->sec3));
#endif
  

  if (conv_h == 1){
    gi_target = &gi_out;
    if ((r = interp_prepare(&gi_in, &gi_out, opt.allow_out_domain, &interp)))
      return r;
  } else {
    gi_target = &gi_in;
  }
  if (conv_v == 1){
    if ((r = z2p_ini(config_v.pnum, &gi_in, opt.udg_miss, 
		     &z2p, config_v.pout))) return r;
  }

  switch (opt.output_format){
  case(FORMAT_GRADS):
    gpv_write_init(&grds, NULL, &gpv);
    break;
  case (FORMAT_GRIB2):
    gpv_write_init(NULL, &grib2_wrt, &gpv);
    break;
  }

  switch(grib2_get_jma_kind(&grib2_in)){
  case (JMA_KIND_ATM): /* meteorological product */

    /* check nz */
    grib2_get_parm("P     ", &parm_cat, &parm_num);
    nz = grib2_inq_nz(&grib2_in, &ymdhm, term_ma, 
		      parm_cat, parm_num);
    org_layer = (double*)malloc(sizeof(double) * nz);
    
    for (kk = 0; kk < nz; kk++){
      org_layer[kk] = (double)(kk + 1);
    }

    if (conv_h == 1 && conv_v == 1){
      if ((r = 
	   gpv_write_pre(JMA_KIND_ATM, 
			 opt.fname_out, &gi_out, opt.pdef_use,
			 EARTH_SHAPE_SPHERE_USERRAD,
			 z2p.pnum, z2p.pout,
			 LAYER_ISOBAR, &ymdhm, 
			 cutoff_time_hour_ma, cutoff_time_min_ma,
			 term_ma,
			 &grib2_in, &gpv))) return r;

      if ((r = gpv_write_convert(&gi_in, nz, 
				 &ymdhm, term_ma, 
				 org_layer, &z2p,
				 &gi_out, &interp,
				 &grib2_in, &gpv))) return r;


    } else if (conv_h == 1) {  
      /* only conv_h == 1, leave vertical cood. unchanged */
      if ((r = gpv_write_pre(JMA_KIND_ATM, 
			     opt.fname_out, &gi_out, opt.pdef_use,
			     EARTH_SHAPE_SPHERE_USERRAD,
			     nz, org_layer,
			     LAYER_HYBRID, &ymdhm, 
			     cutoff_time_hour_ma, cutoff_time_min_ma,
			     term_ma,
			     &grib2_in, &gpv))) return r;
      
      if ((r = gpv_write_convert(&gi_in, nz, 
				 &ymdhm, term_ma, 
				 org_layer, NULL,
				 &gi_out, &interp,
				 &grib2_in, &gpv))) return r;


    } else if (conv_v == 1) { /* only conv_v == 1 */
      if ((r = gpv_write_pre(JMA_KIND_ATM, 
			     opt.fname_out, &gi_in, opt.pdef_use,
			     EARTH_SHAPE_SPHERE_USERRAD,
			     z2p.pnum, z2p.pout,
			     LAYER_ISOBAR, &ymdhm, 
			     cutoff_time_hour_ma, cutoff_time_min_ma,
			     term_ma,
			     &grib2_in, &gpv))) return r;

      if ((r = gpv_write_convert(&gi_in, nz, 
				 &ymdhm, term_ma, 
				 org_layer, &z2p,
				 NULL, NULL,
				 &grib2_in, &gpv))) return r;

    } else { /* just convert to GrADS */
      if ((r = gpv_write_pre(JMA_KIND_ATM, 
			     opt.fname_out, &gi_in, opt.pdef_use,
			     EARTH_SHAPE_SPHERE_USERRAD,
			     nz, org_layer,
			     LAYER_HYBRID, &ymdhm, 
			     cutoff_time_hour_ma, cutoff_time_min_ma,
			     term_ma,
			     &grib2_in, &gpv))) return r;

      if ((r = gpv_write_convert(&gi_in, nz, 
				 &ymdhm, term_ma, 
				 org_layer, NULL,
				 NULL, NULL,
				 &grib2_in, &gpv))) return r;

    }
    break;
  case (JMA_KIND_LAND): /* land */
    nz = 4;  /* ad hoc */
    org_layer = (double*)malloc(sizeof(double) * nz);
    
    org_layer[0] = 0.02;
    org_layer[1] = 0.115;
    org_layer[2] = 0.39;
    org_layer[3] = 0.89;

    if (conv_h == 1){
      if ((r = gpv_write_pre(JMA_KIND_LAND, 
			     opt.fname_out, &gi_out, opt.pdef_use,
			     EARTH_SHAPE_SPHERE_USERRAD,
			     nz, org_layer,
			     LAYER_UGD, &ymdhm, 
			     cutoff_time_hour_ma, cutoff_time_min_ma,
			     term_ma,
			     &grib2_in, &gpv))) return r;

      if ((r = gpv_write_convert_land(&gi_in, nz, 
				      &ymdhm, term_ma, 
				      org_layer,
				      &gi_out, &interp, 
				      &grib2_in, &gpv))) return r;

    } else {
      if ((r = gpv_write_pre(JMA_KIND_LAND, 
			     opt.fname_out, &gi_in, opt.pdef_use,
			     EARTH_SHAPE_SPHERE_USERRAD,
			     nz, org_layer,
			     LAYER_UGD, &ymdhm, 
			     cutoff_time_hour_ma, cutoff_time_min_ma,
			     term_ma,
			     &grib2_in, &gpv))) return r;

      if ((r = gpv_write_convert_land(&gi_in, nz, 
				      &ymdhm, term_ma, 
				      org_layer,
				      NULL, NULL,
				      &grib2_in, &gpv))) return r;

    }
    break;
  case (JMA_KIND_OCEAN): /* Ocean */
    if (conv_h == 1){
      if ((r = gpv_write_pre(JMA_KIND_OCEAN, 
			     opt.fname_out, &gi_out, opt.pdef_use,
			     EARTH_SHAPE_SPHERE_USERRAD,
			     0, NULL,
			     LAYER_SURFACE, &ymdhm, 
			     cutoff_time_hour_ma, cutoff_time_min_ma, 
			     term_ma,
			     &grib2_in, &gpv))) return r;

      if ((r = gpv_write_convert_sst(&gi_in,  
				     &ymdhm, term_ma, 
				     &gi_out, &interp, 
				     &grib2_in, &gpv))) return r;

    } else {
      if ((r = gpv_write_pre(JMA_KIND_OCEAN, 
			     opt.fname_out, &gi_in, opt.pdef_use,
			     EARTH_SHAPE_SPHERE_USERRAD,
			     0, NULL,
			     LAYER_SURFACE, &ymdhm, 
			     cutoff_time_hour_ma, cutoff_time_min_ma, 
			     term_ma,
			     &grib2_in, &gpv))) return r;

      if ((r = gpv_write_convert_sst(&gi_in,  
				     &ymdhm, term_ma, 
				     NULL, NULL,
				     &grib2_in, &gpv))) return r;
    }
    break;
  case(JMA_KIND_RA):
    if (conv_h == 1){

      if ((r = gpv_write_pre(JMA_KIND_RA, 
			     opt.fname_out, &gi_out, opt.pdef_use,
			     EARTH_SHAPE_SPHERE_USERRAD,
			     0, NULL,
			     LAYER_SURFACE, &ymdhm, 
			     cutoff_time_hour_ra, cutoff_time_min_ra, 
			     term_ra,
			     &grib2_in, &gpv))) return r;

      if ((r = gpv_write_convert_ra(&gi_in,  
				    &ymdhm, term_ra, 
				    opt.ra_interp_ope, 
				    &gi_out, &interp,
				    &grib2_in, &gpv))) return r;

    } else {
      if ((r = gpv_write_pre(JMA_KIND_RA, 
			     opt.fname_out, &gi_in, opt.pdef_use,
			     EARTH_SHAPE_GRS80,
			     0, NULL,
			     LAYER_SURFACE, &ymdhm, 
			     cutoff_time_hour_ra, cutoff_time_min_ra, 
			     term_ra,
			     &grib2_in, &gpv))) return r;

      if ((r = gpv_write_convert_ra(&gi_in,  
				    &ymdhm, term_ra, 
				    opt.ra_interp_ope, 
				    NULL, NULL,
				    &grib2_in, &gpv))) return r;
    }
    break;
  default:
    fprintf(stderr, "Error!  Unknown grib2 file.\n");
    return ERR_UNKNOWN_GRIB2;
  }

  gpv_write_post(&gpv);

  free(org_layer);

  if (opt.output_latlon == 1){
    grds_t grds_latlon;

    grds_init(&grds_latlon);

    if ((r = grds_def_latlon(opt.fname_out, 
			     gi_target, opt.pdef_use,
			     &ymdhm, &grds_latlon))) return r;

    if ((r = grds_file_open(&grds_latlon))) return r;

    if ((r = grds_ctl_write(&grds_latlon))) return r;

    if ((r = grds_write_latlon(gi_target,  &grds_latlon))) return r;
    
    grds_file_close(&grds_latlon);

  }

  if (opt.fname_mapfile_in != NULL){
    char fname_map_out[256];

    strcpy(fname_map_out, opt.fname_out);
    strcat(fname_map_out, ".map.dat");

    if ((r = grds_wrt_map(opt.fname_mapfile_in, fname_map_out, 
			  gi_target))) return r;

    
  }

  return 0;
}

