#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "const.h"
#include "z2p.h"

const double tlapse = 0.005;
const double hpa2pa = 100.0;

const double grav = GRAV;
const double rd = RD;

int z2p_ini(int pnum, grid_info_t *gi_ln, unsigned char udg_miss, 
	    z2p_info_t *z2p, double *pout){
  int kz;

  if ((z2p->ip = (N_UI4*)malloc(gi_ln->nxy * pnum *sizeof(int))) == NULL ||
      (z2p->wp0 = (double*)malloc(gi_ln->nxy * pnum *sizeof(double))) == NULL || 
      (z2p->wp1 = (double*)malloc(gi_ln->nxy * pnum *sizeof(double))) == NULL){
    fprintf(stderr, "malloc error at lines %d in %s.\n", __LINE__, __FILE__);
    return ERR_MALLOC;
  }
  z2p->nxy = gi_ln->nxy;
  z2p->pnum = pnum;
  z2p->pout = pout;
  z2p->udg_miss = udg_miss; 
  for (kz = 0; kz < pnum - 1; kz++){
    if (pout[kz] <= pout[kz + 1]){
      fprintf(stderr, 
              "Error! Order of pressures in the pout list is invalid.\n ");
      fprintf(stderr, 
              "Pressures must be set in descending order.\n ");
      return ERR_POUT_ORDER;
    }
  }
  return 0;
}

int z2p_pre(const int nz, const float press[], z2p_info_t *z2p){
  /*
    ik order
      [k * nxy + i]

    input
    pout : pnum
    press : nxy * nz

    output
    ip: nxy * pnum
    wp0, wp1: nxy * pnum

    local variables
    dp0, dp1: nxy * pnum
    logpin: nxy * nz
   */

    int kp;
    int kz;
    int ixy;
    double tmp;
    double *dp0;
    double *dp1;
    double *logpin;
    double *logpout;

    N_UI4 pnum = z2p->pnum;
    N_UI4 nxy = z2p->nxy;
    N_UI4 *ip = z2p->ip;
    double *wp0 = z2p->wp0;
    double *wp1 = z2p->wp1;
    double *pout = z2p->pout;


    if ((logpout = (double*)malloc(pnum * sizeof(double))) == NULL){
      fprintf(stderr, "malloc error at lines %d in %s.\n", __LINE__, __FILE__);
      return ERR_MALLOC;
    }

    for (kp = 0; kp < pnum; kp++){
      logpout[kp] = log(hpa2pa * ((double)pout[kp]));
    }


    if ((logpin = (double*)malloc(nxy * nz * sizeof(double))) == NULL){
      fprintf(stderr, "malloc error at lines %d in %s.\n", __LINE__, __FILE__);
      return ERR_MALLOC;
    }

    for (kz = 0; kz < nz; kz++){
      for (ixy = 0; ixy < nxy; ixy++){
        logpin[kz * nxy + ixy] = log(press[kz * nxy + ixy]);
      }
    }

    if ((dp0 = (double*)malloc(nxy * pnum * sizeof(double))) == NULL || 
	(dp1 = (double*)malloc(nxy * pnum * sizeof(double))) == NULL){
      fprintf(stderr, "malloc error at lines %d in %s.\n", __LINE__, __FILE__);
      return ERR_MALLOC;
    }


    for (ixy = 0; ixy < nxy; ixy++){
      for (kp = 0; kp < pnum; kp++){
        kz = 0;
        while (logpout[kp] < logpin[kz * nxy + ixy]){
	  kz++;
        }
        ip[kp * nxy + ixy] = kz < nz ? kz : nz - 1;
        if (kz != 0){
	  dp0[kp * nxy + ixy] = logpin[(kz- 1) * nxy + ixy] 
	                         - logpout[kp];
        }
        dp1[kp * nxy + ixy] = logpout[kp] - logpin[kz * nxy + ixy];
      }
    }
    for (kp = 0; kp < pnum; kp++){
      for(ixy = 0;  ixy < nxy; ixy++){
        if (ip[kp * nxy + ixy] != 0){
            tmp = 1.0 / (dp0[kp * nxy + ixy] + dp1[kp * nxy + ixy]);
            wp0[kp * nxy + ixy] = dp0[kp * nxy + ixy] * tmp;
            wp1[kp * nxy + ixy] = dp1[kp * nxy + ixy] * tmp;
        }
      }
    }

    free(dp0);
    free(dp1);
    free(logpin);
    free(logpout);
    return 0;

}

int z2p_temp_run(const z2p_info_t *z2p,
		 const float p_surf[], 
                 const float temp_in[], float temp_out[]){

    int kp;
    int ixy;
    double zstrm;
    double zstrp;
    double fact;
    N_UI4 nxy = z2p->nxy;
    N_UI4 pnum = z2p->pnum;
    N_UI4 *ip = z2p->ip;
    double *wp0 = z2p->wp0;
    double *wp1 = z2p->wp1;
    double *pout = z2p->pout;
    
    fact = tlapse * rd / grav;

    for (kp = 0; kp < pnum; kp++){
      for (ixy = 0; ixy < nxy; ixy++){
        if (ip[kp * nxy + ixy] >= 1) {
	  zstrm = temp_in[(ip[kp * nxy + ixy] - 1) * nxy + ixy];
	  zstrp = temp_in[ip[kp * nxy + ixy] * nxy + ixy];
	  temp_out[kp * nxy + ixy] = zstrm * wp1[kp * nxy + ixy] 
	                           + zstrp * wp0[kp * nxy + ixy];
        } else if (z2p->udg_miss == 0){
	  zstrp = temp_in[0 * nxy + ixy];
	  temp_out[kp * nxy + ixy] = zstrp 
	    * pow(hpa2pa * (double)pout[kp] /  p_surf[ixy], fact);
        } else { /* udg_miss == 1 */
	  temp_out[kp * nxy + ixy] = MISSING;
	}
      }
    }
    return 0;
}

int z2p_gph_run(const z2p_info_t *z2p, const float zs[], const float z_f[], 
		const float temp_surf[],  
		const float temp_out[], 
		float gph_out[]){
    
  int kp;
  int ixy;
  double zstrm;
  double zstrp;
  N_UI4 nxy = z2p->nxy;
  N_UI4 pnum = z2p->pnum;
  N_UI4 *ip = z2p->ip;
  double *wp0 = z2p->wp0;
  double *wp1 = z2p->wp1;

  double r_tlapse = 1.0 / tlapse;

  
  for (kp = 0; kp < pnum; kp++){
    for (ixy = 0; ixy < nxy; ixy++){
      if (ip[kp * nxy + ixy] >= 1) {
	zstrp = z_f[ip[kp * nxy + ixy] * nxy + ixy];
	zstrm = z_f[(ip[kp * nxy + ixy] - 1) * nxy + ixy];
	gph_out[kp * nxy + ixy] = (zstrm * wp1[kp * nxy + ixy] 
			                   + zstrp * wp0[kp * nxy + ixy]);
      } else if (z2p->udg_miss == 0){
	zstrp = z_f[0 * nxy + ixy];
	gph_out[kp * nxy + ixy] = zstrp 
	  - (temp_out[kp * nxy + ixy] - temp_surf[ixy]) * r_tlapse;
      } else { /* udg_miss == 1 */
	gph_out[kp * nxy + ixy] = MISSING;
      }
    }
  }
  return 0;
}

int z2p_gen_run(const z2p_info_t *z2p, 
                const float fld_in[], float fld_out[]){
  
  int kp;
  int ixy;
  double zstrm;
  double zstrp;
  N_UI4 nxy = z2p->nxy;
  N_UI4 pnum = z2p->pnum;
  N_UI4 *ip = z2p->ip;
  double *wp0 = z2p->wp0;
  double *wp1 = z2p->wp1;

  for (kp = 0; kp < pnum; kp++){
    for (ixy = 0; ixy < nxy; ixy++){
      if (ip[kp * nxy + ixy] >= 1) {
	zstrp = fld_in[ip[kp * nxy + ixy] * nxy + ixy];
	zstrm = fld_in[(ip[kp * nxy + ixy] - 1) * nxy + ixy];
	fld_out[kp * nxy + ixy] = zstrm * wp1[kp * nxy + ixy] 
	                        + zstrp * wp0[kp * nxy + ixy];
      } else if (z2p->udg_miss == 0){
	fld_out[kp * nxy + ixy] = fld_in[0 * nxy + ixy];
      } else {
	fld_out[kp * nxy + ixy] = MISSING;
      }
    }
  }
  return 0;
}
